/*
 * SpiceStream - simple, incremental reader for analog data files,
 * such as those produced by spice-type simulators.
 * 
 * include LICENSE
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include <stdarg.h>
#include <errno.h>

#include <glib.h>

#include <msglog.h>
#include <spicestream.h>
#include <regex.h>
#define REGEXP_T regex_t
#define regexp_test(c, s) (regexec((c), (s), 0, NULL, 0) == 0)


extern SpiceStream *sf_rdhdr_hspice( FILE *fp);
extern SpiceStream *sf_rdhdr_hsascii( FILE *fp);
extern SpiceStream *sf_rdhdr_hsbin( FILE *fp);
extern SpiceStream *sf_rdhdr_cazm( FILE *fp);
extern SpiceStream *sf_rdhdr_s3raw( FILE *fp);
extern SpiceStream *sf_rdhdr_s2raw( FILE *fp);
extern SpiceStream *sf_rdhdr_ascii( FILE *fp);
extern SpiceStream *sf_rdhdr_nsout( FILE *fp);

static int ss_readrow_none(SpiceStream *, double *ivar, double *dvars);


typedef SpiceStream * (*ReadHeaderFP) (FILE *fp);
typedef struct _FileFormat FileFormat;
   
struct _FileFormat {
   char *name;
   char *fnrexp;
   REGEXP_T *creg;/* compiled form of regexp */
   ReadHeaderFP rdfunc;
};

static FileFormat format_tab[] = {
   { "hspice",  "\\.(tr|sw|ac)[0-9]$", NULL,
	 sf_rdhdr_hspice },
   { "hsascii",  "\\.(tr|sw|ac)[0-9]$", NULL, 
	 sf_rdhdr_hsascii },
   { "hsbinary",  "\\.(tr|sw|ac)[0-9]$", NULL,
	 sf_rdhdr_hsbin },
   { "cazm",  "\\.[BNW]$",  NULL,
	 sf_rdhdr_cazm },
   { "spice3raw", "\\.raw$", NULL,
	 sf_rdhdr_s3raw },
   { "spice2raw",  "\\.rawspice$", NULL,
	 sf_rdhdr_s2raw },
   { "ascii",  "\\.(asc|acs|ascii)$", NULL,
	 sf_rdhdr_ascii },
   { "nsout",  "\\.out$" , NULL,
	 sf_rdhdr_nsout },
};
static const int NFormats = sizeof(format_tab) / sizeof(format_tab[0]);


#define SS_NOT_USE_REGEX 0
#define SS_USE_REGEX 1

static regex_t *
regexp_compile(char *str)
{
   int err;
   char ebuf[128];
   regex_t *creg;

   creg = g_new(regex_t, 1);
   err = regcomp(creg, str, REG_NOSUB | REG_EXTENDED);
   if (err) {
      regerror(err, creg, ebuf, sizeof(ebuf));
      msg_fatal( "internal error (in regexp %s):\n  %s", str, ebuf );
   }
   return creg;
}

/*
 * Open spice waveform file for reading.  
 * Reads in header with signal names (and sometimes signal types).
 * TODO: simple strategies for trying to deduce file type from
 * name or contents.
 */

SpiceStream *
ss_find_format( FILE *fp, char *filename , int use_regex, int *index)
{      
   int i;
   SpiceStream *ss;
   
   for (i = 0; i < NFormats; i++) {
      if ( ! format_tab[i].creg) {
	 format_tab[i].creg = regexp_compile(format_tab[i].fnrexp);
      }
      if (use_regex == 0 || regexp_test(format_tab[i].creg, filename)) {
	 ss = (format_tab[i].rdfunc)(fp);
	 if (ss) {
	    *index = i;
	    msg_dbg("%s: read with format \"%s\"", filename, format_tab[i].name);
	    return ss;
	 }
	 
	 if (fseeko(fp, 0L, SEEK_SET) < 0) {
	    msg_fatal("Seek error '%s': %s", filename, strerror(errno) );
	 }
      }
   }
   return NULL;
}

int
ss_format_index(char *format)
{
   int i;
   
   for (i = 0; i < NFormats; i++) {
      if ( strcmp(format, format_tab[i].name) == 0) {
	 return i;
      }
   }
   return -1;
}
      
SpiceStream *
ss_open(char *filename, char *format)
{
   FILE *fp;
   int i;
   SpiceStream *ss;
   
   fp = fopen(filename, "r"); /* fopen64 */
   if (fp == NULL) {
      msg_error( "fopen(\"%s\"): %s\n", filename, strerror(errno));
      return NULL;
   }
   if (format == NULL) {
      /* try to guess format from filename */
      ss = ss_find_format( fp, filename, SS_USE_REGEX, &i );
      if ( ! ss) {
	 msg_dbg( "%s: couldn't guess a format from filename suffix.", filename);
	 ss = ss_find_format( fp, filename, SS_NOT_USE_REGEX, &i );
	 if ( ! ss) {
	    msg_error("%s: couldn't read with any format\n", filename);
	    return NULL;
	 }
      }
   } else {
      if ( ( i = ss_format_index(format)) < 0 ) {
	 msg_error( "Format \"%s\" unknown", format);
	 return NULL;
      } else {
	 ss = (format_tab[i].rdfunc)(fp);
	 if ( ! ss) {
	    msg_error("failed to open %s using format %s", filename, format);
	    return NULL;
	 }
      }
   }
   ss->filetype = i;
   ss->filename = g_strdup(filename);
   return ss;
}


/*
 * Allocate SpiceStream structure and fill in some portions.
 * To be called only from format-specific header-reading functions,
 * usually after they read and verify the header.
 * Caller must still set types and names of ivar and dvars,
 * and must set readrow and linebuf items.
 */
SpiceStream *
ss_new(FILE *fp, int ndv, int nspar)
{
   SpiceStream *ss;

   ss = g_new0(SpiceStream, 1);
   ss->fp = fp;
   ss->ivar = g_new0(SpiceVar, 1);
   ss->ndv = ndv;
   if (ndv) {
      ss->dvar = g_new0(SpiceVar, ndv);
   }
   ss->nsweepparam = nspar;
   if (nspar) {
      ss->spar = g_new0(SpiceVar, nspar);
   }

   return ss;
}

/*
 * Close the file assocated with a SpiceStream.
 * No more data can be read, but the header information can still
 * be accessed.
 */
void ss_close(SpiceStream *ss)
{
   fclose(ss->fp);
   ss->fp = NULL;
   ss->readrow = ss_readrow_none;
}

/*
 * Free all resources associated with a SpiceStream.
 */
void ss_delete(SpiceStream *ss)
{
   if (ss->fp)
      fclose(ss->fp);
   if (ss->filename)
      g_free(ss->filename);
   if (ss->ivar)
      g_free(ss->ivar);
   if (ss->dvar)
      g_free(ss->dvar);
   if (ss->linebuf)
      g_free(ss->linebuf);
   g_free(ss);
}

/*
 * row-reading function that always returns EOF.
 */
static int 
ss_readrow_none(SpiceStream *ss, double *ivar, double *dvars)
{
   return 0;
}


static char *vartype_names[] = {
   "Unknown", "Time", "Voltage", "Current", "Frequency"
};
const int nvartype_names = sizeof(vartype_names) / sizeof(char *);

/*
 * return a string corresponding to a SpiceStream VarType.
 * the pointer returned is in static or readonly storage,
 * and is overwritten with each call.
 */
char *vartype_name_str(VarType type)
{
   static char buf[32];
   if (type < nvartype_names){
      return vartype_names[type];
   } else {
      sprintf(buf, "type-%d", type);
      return buf;
   }
}

/*
 * return pointer to string with printable name for a variable
 * or one of the columns of a variable.
 * buf is a pointer to a buffer to use.  If NULL, one will be allocated.
 * n is the maximum number of characters to put in the buffer.
 */
char *ss_var_name(SpiceVar *sv, int col, char *buf, int n)
{
   int idx;

   if (buf == NULL) {
      int l;
      l = strlen(sv->name + 3);
      buf = g_new(char, l);
      n = l;
   }
   strncpy(buf, sv->name, n-1);
   n -= strlen(buf)+1;
   if(sv->ncols == 1 || col < 0)
      return buf;
   if (n>1) {
      idx = strlen(buf);
      buf[idx++] = '.';
      buf[idx++] = '0' + col;
      buf[idx] = 0;
   }
	
   return(buf);
}

/*
 * given a filetype number, return a pointer to a string containing the 
 * name of the Spicestream file format.
 * Valid file type numbers start at 0. 
 */
char *ss_filetype_name(int n)
{
   if (n >= 0 && n < NFormats) {
      return format_tab[n].name;
   } else {
      return NULL;
   }
}

/*
 * utility function to read whole line into buffer, expanding buffer if needed.
 * line buffer must be allocated with g_malloc/g_new, or NULL in which case
 * we allocate an initial, buffer.
 * returns 0 or EOF.
 */
int
fread_line(FILE *fp, char **bufp, int *bufsize)
{
   int c;
   int n = 0;
   if (*bufp == NULL) {
      if (*bufsize == 0)
	 *bufsize = 1024;
      *bufp = g_new(char, *bufsize);
   }
   while(((c = getc(fp)) != EOF) && c != '\n') {
      (*bufp)[n++] = c;
      if (n >= *bufsize) {
	 *bufsize *= 2;
	 *bufp = g_realloc(*bufp, *bufsize);
      }
   }
   (*bufp)[n] = 0;
   if ( n > 0 && (*bufp)[n - 1] == '\r' ){
      (*bufp)[n - 1] = 0;
   }
   if (c == EOF){
      return EOF;
   } else {
      return 0;
   }
}


