/*
 * ss_spice3.c: routines for SpiceStream that handle the file formats
 * 	known as Berkeley Spice3 Rawfile
 * 
 * include LICENSE
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <float.h>

#include <glib.h>
#include <msglog.h>
#include <spicestream.h>

static int sf_readrow_s3raw(SpiceStream *sf, double *ivar, double *dvars);
char *msgid = "s3raw";
static int sf_readrow_s3bin(SpiceStream *sf, double *ivar, double *dvars);

/* convert variable type string from spice3 raw file to 
 * our type numbers
 */
static VarType
sf_str2type_s3raw(char *s)
{
   if (strcasecmp(s, "voltage") == 0)
      return VOLTAGE;
   else if (strcasecmp(s, "current") == 0)
      return CURRENT;
   else if (strcasecmp(s, "frequency") == 0)
      return FREQUENCY;
   else if (strcasecmp(s, "time") == 0)
      return TIME;
   else return UNKNOWN;
}


/* Read spice-type file header - Berkeley Spice3 "raw" format */
SpiceStream *
sf_rdhdr_s3raw( FILE *fp)
{
   SpiceStream *sf = NULL;
   char *line = NULL;
   int lineno = 0;
   int linesize = 1024;
   char *key, *val;
   int nvars = 0;
   int npoints = 0;
   int got_nvars = 0;
   int got_values = 0;
   int dtype_complex = 0;
   int binary = 0;
   char *vnum, *vname, *vtypestr;
   int i;
   
   while (fread_line(fp, &line, &linesize) != EOF) {
      lineno++;
      if(lineno == 1 && strncmp(line, "Title: ", 7)) {
	 /* not a spice3raw file; bail out */
	 msg_dbg("%d: Doesn't look like a spice3raw file; \"Title:\" expected\n",
		 lineno);
	 return NULL;
      }
      
      key = strtok(line, ":");
      if ( ! key) {
	 msg_error("%d: syntax error, expected \"keyword:\"", lineno);
	 g_free(line);
	 return NULL;
      }
      if (strcmp(key, "Flags") == 0) {
	 while ((val = strtok(NULL, " ,\t\n"))) {
	    if (strcmp(val, "real") == 0) {
	       dtype_complex = 0;
	    }
	    if (strcmp(val, "complex") == 0) {
	       dtype_complex = 1;
	    }
	 }
      } else if (strcmp(key, "No. Variables") == 0) {
	 val = strtok(NULL, " \t\n");
	 if ( ! val) {
	    msg_error("%d: syntax error, expected integer", lineno);
	    g_free(line);
	    return NULL;
	 }
	 nvars = atoi(val);
	 got_nvars = 1;
      } else if (strcmp(key, "No. Points") == 0) {
	 val = strtok(NULL, " \t\n");
	 if ( ! val) {
	    msg_error("%d: syntax error, expected integer", lineno);
	    g_free(line);
	    return NULL;
	 }
	 npoints = atoi(val);
      } else if(strcmp(key, "Variables") == 0) {
	 if (!got_nvars) {
	    msg_error( "%d: \"Variables:\" before \"No. Variables:\"", lineno );
	    goto err;
	 }
	 sf = ss_new(fp, nvars - 1, 0);
	 sf->ncols = 1;
	 sf->ntables = 1;
	 /* first variable may be described on the same line
	  * as "Variables:" keyword
	  */
	 vnum = strtok(NULL, " \t\n");
	 
	 for (i = 0; i < nvars; i++) {
	    if (i || !vnum) {
	       if (fread_line(fp, &line, &linesize) == EOF) {
		  msg_error( "%d: Unexpected EOF in \"Variables:\" at var %d",
			     lineno, i);
		  goto err;
	       }
	       lineno++;
	       vnum = strtok(line, " \t\n");
	    }
	    vname = strtok(NULL, " \t\n");
	    vtypestr = strtok(NULL, " \t\n");
	    if (!vnum || !vname || !vtypestr) {
	       msg_error("%d: expected number name type", lineno);
	       goto err;
	    }
	    if (i == 0) { /* assume Ind.Var. first */
	       sf->ivar->name = g_strdup(vname);
	       sf->ivar->type = sf_str2type_s3raw(vtypestr);
	       sf->ivar->col = 0;
	       /* ivar can't really be two-column,
		this is a flag that says to
		discard 2nd point */
	       if (dtype_complex)
		  sf->ivar->ncols = 2;
	       else
		  sf->ivar->ncols = 1;
	       
	    } else {
	       sf->dvar[i-1].name = g_strdup(vname);
	       sf->dvar[i-1].type = sf_str2type_s3raw(vtypestr);
	       sf->dvar[i-1].col = sf->ncols;
	       if (dtype_complex)
		  sf->dvar[i-1].ncols = 2;
	       else
		  sf->dvar[i-1].ncols = 1;
	       
	       sf->ncols += sf->dvar[i-1].ncols;
	    }
	 }
      } else if (strcmp(key, "Values") == 0) {
	 got_values = 1;
	 break;
      } else if (strcmp(key, "Binary") == 0) {
	 binary = 1;
	 got_values = 1;
	 break;
      }
      if (got_values)
	 break;
   }
   if ( ! sf) {
      msg_error("%d: no \"Variables:\" section in header", lineno);
      goto err;
   }
   if ( ! got_values) {
      msg_error( "%d: EOF without \"Values:\" in header", lineno);
      goto err;
   }

   if (binary) {
      sf->readrow = sf_readrow_s3bin;
   } else {
      sf->readrow = sf_readrow_s3raw;
   }
   sf->read_rows = 0;
   sf->expected_vals = npoints * (sf->ncols + (dtype_complex ? 1 : 0));
   msg_dbg("expecting %d values\n", sf->expected_vals);
   sf->lineno = lineno;
   sf->linebuf = line;
   sf->lbufsize = linesize;
   msg_dbg("Done with header at offset 0x%lx\n", (long) ftello(sf->fp));
	
   return sf;
err:
   if (line)
      g_free(line);
   if (sf) {
      sf->fp = NULL;  
      /* prevent ss_delete from cleaning up FILE*; ss_open callers 
       may rewind and try another format on failure. */
      ss_delete(sf);
   }
   return NULL;
}

/*
 * Read row of values from an ascii spice3 raw file
 */
static int
sf_readrow_s3raw(SpiceStream *sf, double *ivar, double *dvars)
{
	int i;
	int frownum;
	char *tok;
	double v;

   if ((sf->flags & SSF_PUSHBACK) == 0) {
      do {
	 if (fread_line(sf->fp, &sf->linebuf, &sf->lbufsize) == EOF) {
	    return 0;  /* normal EOF */
	 }
	 sf->lineno++;
	 /* first line of a set contains row number and independent variable.
	  * there may be a blank line between line sets
	  */
	 tok = strtok(sf->linebuf, " \t\n,");
      } while(!tok);
      if (!tok) {
	 msg_error("%s:%d: expected row number", 
		   sf->filename, sf->lineno);
	 return -1;
      }
      if (!isdigit(*tok)) {
	 msg_warning( "%s:%d: expected row number, got \"%s\". Note: only one dataset per file is supported, extra garbage ignored", 
		      sf->filename, sf->lineno, tok);
	 return 0;
      }
      frownum = atoi(tok);
      /* todo: check for expected and maximum row number */

      tok = strtok(NULL, " \t\n,");
      if (!tok) {
	 msg_warning("%s:%d: expected ivar value", 
		     sf->filename, sf->lineno);
	 return -1;
      }
      v = g_ascii_strtod(tok, NULL);
      if (v < sf->ivval) {
	 /* independent-variable value decreased, this must
	  * be the start of another sweep.  hold the value and
	  * return flag to caller.
	  */
	 sf->ivval = v;
	 sf->flags |= SSF_PUSHBACK;
	 return -2;
      } else {
	 sf->ivval = v;
	 *ivar = v;
      }
   } else {
      /* iv value for start of new sweep was read last time */
      sf->flags &= ~SSF_PUSHBACK;
      *ivar = sf->ivval;
   }
	
   for (i = 0; i < sf->ndv; i++) {
      SpiceVar *dv;
      dv = &sf->dvar[i];
      
      if (fread_line(sf->fp, &sf->linebuf, &sf->lbufsize) == EOF) {
	 msg_error("%s:%d unexpected EOF at dvar %d",
		   sf->filename, sf->lineno, i);
	 return -1;
      }
      sf->lineno++;

      tok = strtok(sf->linebuf, " \t\n,");
      if (!tok) {
	 msg_error("%s:%d: expected value", 
		   sf->filename, sf->lineno);
	 return -1;
      }
      dvars[dv->col-1] = g_ascii_strtod(tok, NULL);

      if (dv->ncols > 1) {
	 tok = strtok(NULL, " \t\n,");
	 if (!tok) {
	    msg_error("%s:%d: expected second value", 
		   sf->filename, sf->lineno);
	    return -1;
	 }
	 dvars[dv->col] = g_ascii_strtod(tok, NULL);
      }
   }
   sf->read_rows++;
   return 1;
}

/*
 * Read a single value from binary spice3 rawfile, and do
 * the related error-checking.
 */

static int
sf_getval_s3bin(SpiceStream *sf, double *dval)
{
   off_t pos; /* off64_t */
   double val;

   if (sf->read_vals >= sf->expected_vals) {
      pos = ftello(sf->fp);
      msg_dbg("past last expected value offset 0x%lx", (long) pos);
      return 0;
   }
   if (fread(&val, sizeof(double), 1, sf->fp) != 1) {
      pos = ftello(sf->fp);
      msg_error("unexepected EOF in data at offset 0x%lx", (long) pos);
      return -1;
   }
   sf->read_vals++;

   *dval = val;
   return 1;
}


/*
 * Read row of values from a binay spice3 raw file
 */
static int
sf_readrow_s3bin(SpiceStream *sf, double *ivar, double *dvars)
{
   int i, rc;
   double v;
   double dummy;
	
   if ((sf->flags & SSF_PUSHBACK) == 0) {
      rc = sf_getval_s3bin(sf, &v);
      if (rc == 0)		/* file EOF */
	 return 0;
      if (rc < 0)
	 return -1;
      if (sf->ivar->ncols == 2) {
	 rc = sf_getval_s3bin(sf, &dummy);
	 if(rc == 0)		/* file EOF */
	    return 0;
	 if (rc < 0)
	    return -1;
      }
      if (v < sf->ivval) {
	 /* independent-variable value decreased, this must
	  * be the start of another sweep.  hold the value and
	  * return flag to caller.
	  */
	 sf->ivval = v;
	 sf->flags |= SSF_PUSHBACK;
	 return -2;
      } else {
	 sf->ivval = v;
	 *ivar = v;
      }
   } else {
      /* iv value for start of new sweep was read last time */
      sf->flags &= ~SSF_PUSHBACK;
      *ivar = sf->ivval;
   }

   for (i = 0; i < sf->ncols-1; i++) {
      if (sf_getval_s3bin(sf, &dvars[i]) != 1) {
	 msg_warning("%s: EOF or error reading data field %d in row %d; file is incomplete.",
		     sf->filename, i, sf->read_rows);
	 return 0;
      }
   }

   sf->read_rows++;
   return 1;
}
