/*
 * gawapp.c - gaw application stuff
 * 
 * include LICENSE
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <gaw.h>

#include <gawpixmaps.h>

#define GAW_PRIVATE_DND_MAGIC 0xf00bbaad

enum {
  TARGET_STRING,
  TARGET_DVAR,
  TARGET_ROOTWIN
};

static GtkTargetEntry target_table[] = {
  { "x-gaw/dvar", GTK_TARGET_SAME_APP, TARGET_DVAR },
  { "STRING",     0, TARGET_STRING },
  { "text/plain", 0, TARGET_STRING },
  { "application/x-rootwindow-drop", 0, TARGET_ROOTWIN }
};

static guint n_targets = sizeof(target_table) / sizeof(target_table[0]);

/*
 *  prototypes
 */

static void ap_update_xcursor(WavePanel *wp, AWCursor *csp, int x);


void
ap_set_drag_dest(GtkWidget *widget, WavePanel *wp)
{
   gtk_drag_dest_set ( widget,
		      GTK_DEST_DEFAULT_ALL,
		      target_table, n_targets - 1, /* no rootwin */
		      GDK_ACTION_COPY | GDK_ACTION_MOVE);
   g_signal_connect (widget, "drag_data_received",
		     G_CALLBACK (ap_dnd_target_event), wp);
}

/* 
 * Set the X pointer cursor for all wavepanels: used to provide a
 * hint that we're expecting the user to drag out a line or region.
 */


void
ap_set_all_wp_cursors(UserData *ud, int cnum)
{
   GdkCursor *cursor;
   int i;
   WavePanel *wp;

   if (cnum == -1) {
      cursor = NULL;
   } else {
      cursor = gdk_cursor_new(cnum);
   }
   for (i = 0; i < ud->npanels; i++) {
      wp = ud->panels[i];
      gdk_window_set_cursor(wp->drawing->window, cursor);
   }
   if (cursor) {
      gdk_cursor_destroy(cursor);
   }
}

void 
ap_widget_show(GtkWidget *w,  gboolean show)
{
   if ( show) {
      if ( ! GTK_WIDGET_VISIBLE(w)) {
	 gtk_widget_show(w);
      }
   } else {
      if ( GTK_WIDGET_VISIBLE(w)) {
	 gtk_widget_hide(w);
      }
   }
}

void
ap_draw_panel_ylabels(WavePanel *wp)
{
   if (wp->label_min) {
      gtk_label_set(GTK_LABEL(wp->label_min),
		    val2str(wp->start_yval, CNV_NORMAL));
   }
   if(wp->label_max) {
      gtk_label_set(GTK_LABEL(wp->label_max),
		    val2str(wp->end_yval, CNV_NORMAL));
   }
}

static GtkWidget * 
ap_create_toggle_button(GtkWidget *box, gchar *brc, gchar *tips,
			 GtkTooltips *tooltips )
{
   GtkWidget *button;
   
   button = gtk_toggle_button_new ();
   if ( box ) {
      gtk_box_pack_end(GTK_BOX(box), button,  FALSE, FALSE, 0);
   }
   gtk_widget_set_name(button, brc );
   gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips), button, tips, "");
   gtk_widget_show(button); /* show the button to get the color !!! */
   
   return button;
}

static GtkWidget * 
ap_create_measure_button(GtkWidget *box, gchar *brc, gchar *tips,
			 GtkTooltips *tooltips )
{
   GtkWidget *button;
   
   button = gtk_button_new ();
   if ( box ) {
      gtk_box_pack_end(GTK_BOX(box), button,  FALSE, FALSE, 0);
   }
   gtk_widget_set_name(button, brc );
   gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips), button, tips, "");
   gtk_widget_show(button); /* show the button to get the color !!! */
   
   return button;
}

static GtkWidget * 
ap_create_measure_label(GtkWidget *box, gchar *blabel,  gchar *lrc )
{
   GtkWidget *label;
   
   label = gtk_label_new (blabel);
   gtk_widget_set_name(label, lrc );
   gtk_widget_show(label);
   gtk_container_add( GTK_CONTAINER(box), label );
   
   return label;
}

static void
ap_create_measure_unit(UserData *ud)
{
   GtkWidget *hbox;
   GtkWidget *button;
   GtkWidget *label;
   gchar text[80];
   gchar tips[80];
   int i;
   
   /* hbox for 3 cursor labels */
   hbox = gtk_hbox_new(FALSE, 0);
   gtk_widget_show(hbox);
   gtk_box_pack_start(GTK_BOX(ud->mainBox), hbox, FALSE, FALSE, 0);

   if ( ! ud->cursors ) {
      ud->cursors = g_new0(AWCursor*, 3);
   }

   /* create the 3 X buttons  */
   for ( i = AW_NX_MBTN - 1 ; i >= 0 ; i-- ) {
      ud->cursors[i] = g_new0(AWCursor, 1);
      if ( i == 2 ) {
	   /* diff button */
	 sprintf(text, "cursorD");
	 sprintf(tips, "measure: cursor1 - cursor0");
      } else {
	 sprintf(text, "cursor%d", i);
	 sprintf(tips, "measure: cursor%d", i);
      }
      button = ap_create_measure_button(hbox, "wavebutton", 
			tips, ud->tooltips );
      label = ap_create_measure_label(button, "0.0", text ); 
     
      ud->cursors[i]->button = button;
      ud->cursors[i]->label = label;
      ud->cursors[i]->color_name = g_strdup(text);
   }
}

static void
ap_create_xlabel_box(UserData *ud)
{
   GtkWidget *hbox;
   GtkWidget *label;
   
   hbox = gtk_hbox_new(FALSE, 0);
   ud->xlabel_box = hbox ;
   g_object_ref (hbox); /* increment ref to avoid destruction */
   
   label = gtk_label_new("0");
   gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   ud->win_xlabel_left = label;

   label = gtk_label_new("0");
   gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   ud->win_xlabel_right = label;

   label = gtk_label_new("LogX");
   gtk_box_pack_end(GTK_BOX(hbox), label, TRUE, FALSE, 0);
   ud->win_xlabel_log = label;
   ap_widget_show(label, userPrefs->setLogX);
   
   if (userPrefs->showXLabels) {
      ap_widget_show(hbox, TRUE);
   } else {
      ap_widget_show(hbox, FALSE);
   }
   return;
}

/*
 * update text labeling the waveform graphs' X-axis
 */
void
ap_draw_xlabels(UserData *ud)
{
   gtk_label_set(GTK_LABEL(ud->win_xlabel_left),
		 val2str(ud->start_xval, CNV_NORMAL));
   gtk_label_set(GTK_LABEL(ud->win_xlabel_right),
		 val2str(ud->end_xval, CNV_NORMAL));
}

/*
 * main win horizontal scroolbar handler
 */
void
win_scroll_handler_cb(GtkWidget *widget, gpointer data )
{
   GtkAdjustment *adj = GTK_ADJUSTMENT(widget);
   int i;
   WavePanel *wp;
   UserData *ud = (UserData *) data;

   msg_dbg(" called");
   if ( ! userPrefs->setLogX ) {
      ud->start_xval = adj->value * ( ud->max_xval - ud->min_xval ) +
	 ud->min_xval;
      ud->end_xval = (adj->value + adj->page_size ) *
	 ( ud->max_xval - ud->min_xval ) + ud->min_xval;
   } else {
      ud->start_xval = ud->min_xval * pow( ud->max_xval /
					   ud->min_xval, adj->value );
      ud->end_xval   = ud->min_xval * pow( ud->max_xval / ud->min_xval,
					   adj->value + adj->page_size );
   }
   ap_draw_xlabels(ud);

   for (i = 0; i < ud->npanels ; i++) {
      AWCursor *csp;
      wp = ud->panels[i];
      wp->start_xval = ud->start_xval;
      wp->end_xval = ud->end_xval;

      csp = ud->cursors[0];
      if ( csp->shown ) {
	 ap_update_xcursor(wp, csp, csp->x);
      }
      csp = ud->cursors[1];
      if ( csp->shown ) {
	 ap_update_xcursor(wp, csp, csp->x);
      }
      
      if ( ud->suppress_redraw == 0) {
	 ap_draw_wavepanel(wp->drawing, NULL, wp);
      }
   }
}

/*
 * main win bottom right labels and scrollbar
 */
static void
ap_create_win_bottom(UserData *ud)
{
   GtkAdjustment *adj;
   GtkWidget *scrollbar;
   
   /* horizontal box for X-axis labels */
   ap_create_xlabel_box(ud);

   /* scrollbar */
   adj = (GtkAdjustment *)
      gtk_adjustment_new(0.0, /* value */
			 0.0, /* lower */
			 1.0, /* upper */
			 1.0/100.,    /* step increment = 1% */
			 1.0/2.,      /* page increment = 50% */
			 1.0          /* page_size */
			 );
   ud->xadj = adj;
   scrollbar = gtk_hscrollbar_new(GTK_ADJUSTMENT(adj));
   gtk_range_set_update_policy (GTK_RANGE (scrollbar), 
				GTK_UPDATE_CONTINUOUS);
   g_signal_connect( adj, "value_changed", 
		     G_CALLBACK (win_scroll_handler_cb), (gpointer) ud );
   gtk_widget_show(scrollbar);
   ud->xscrollbar = scrollbar;
   g_object_ref (scrollbar); /* increment ref to avoid destruction */
}


void
ap_mbtn_update(MeasureBtn *mbtn, UserData *ud )
{
   double mvalue;
   gint show;
   
   if ( mbtn->csp == ud->cursors[2] ) {
      /* Y diff button */
      double y0 = wv_interp_value(mbtn->var, ud->cursors[0]->xval);
      double y1 = wv_interp_value(mbtn->var, ud->cursors[1]->xval);
      mvalue = y1 - y0;
      show = mbtn->csp->shown & userPrefs->showYDiff;
   } else {
      mvalue = wv_interp_value(mbtn->var, mbtn->csp->xval);
      show = mbtn->csp->shown ;
   }
   
   if (show) {
      gtk_label_set(GTK_LABEL(mbtn->label), val2str(mvalue, CNV_NORMAL));
   }
   ap_widget_show(mbtn->button, show);
}


void
ap_mbtn_update_all(UserData *ud)
{
   msg_dbg("called");
   g_list_foreach(ud->all_measure_buttons, (GFunc) ap_mbtn_update, (gpointer) ud);
}


/*
 * simply forces widget w to redraw itself
 *    indirect call to expose event
 */
void
ap_panel_redraw(WavePanel *wp)
{
   GtkWidget *w = wp->drawing;
   GdkRectangle rect;
   
   rect.x = 0;
   rect.y = 0;
   rect.width = w->allocation.width;
   rect.height = w->allocation.height;

   if (! GTK_WIDGET_REALIZED(w)) {
      return;
   }
   /* ask redraw */
   gdk_window_invalidate_rect (w->window, &rect, FALSE);
}

void
ap_panel_full_redraw(WavePanel *wp)
{
   /* update max and min values */
   am_wavepanel_update_data(wp);
   am_wavetable_update_data(wp->ud);

   ap_panel_redraw(wp);
   ap_draw_panel_ylabels(wp);
}

void
ap_all_panel_redraw(UserData *ud)
{
   WavePanel *wp;
   int i;
   
   for (i = 0 ; i < ud->npanels ; i++) {
      wp = ud->panels[i];
      ap_panel_redraw(wp);
   }
}

static void
ap_update_panel_selected( UserData *ud, guint state)
{
   int i;

   if ( state & GDK_SHIFT_MASK ){
      /* if shift key, don't unselects panel */
      return;
   }
   for (i = 0 ; i < ud->npanels ; i++) {
      WavePanel *wp = ud->panels[i];
      
      wp->selected = 0;  /* unselect all panels */
   }
}

/*
 * draw (or undraw) a vertical-bar cursor in all panels.
 */

static void
ap_draw_cursor(AWCursor *csp, UserData *ud)
{
   int h, i;
   WavePanel *wp;
   
   for (i = 0 ; i < ud->npanels ; i++) {
      wp = ud->panels[i];
      
      h = wp->drawing->allocation.height;
      
      if (wp->drawing->window){
	 gdk_draw_line(wp->drawing->window, csp->cursor_gc,
		       csp->x, 0, csp->x, h);
      }
   }
}

static void
ap_update_xcursor(WavePanel *wp, AWCursor *csp, int x)
{
   UserData  *ud = wp->ud;
   double xval;
   AWCursor *csp1;

   xval = x2val(wp, x, userPrefs->setLogX);
   if (xval < wp->start_xval || xval > wp->end_xval) {
      return;
   }
   /*
    * move cursor at specified location;
    * turn it on if not on already.
    */
   if (csp->shown) {
      /* undraw old cursor */
      ap_draw_cursor(csp, ud);
   } else {
      ap_widget_show(csp->button, TRUE);
   }

   csp->x = x;
   csp->xval = xval;
   csp->shown = 1;
   /* draw cursor in each panel */
   ap_draw_cursor(csp, ud);
   
   /* show label */
   gtk_label_set(GTK_LABEL(csp->label), val2str(csp->xval, CNV_NORMAL) );

   csp = ud->cursors[0];
   csp1 = ud->cursors[1];
   if (  csp->shown && csp1->shown ) {
      if ( ! ud->cursors[2]->shown ) {
	 ap_widget_show(ud->cursors[2]->button, TRUE);
	 ud->cursors[2]->shown = 1;
      }
      ud->cursors[2]->x = csp1->x -  csp->x;
      ud->cursors[2]->xval = csp1->xval -  csp->xval;
      gtk_label_set(GTK_LABEL(ud->cursors[2]->label),
		    val2str(ud->cursors[2]->xval, CNV_NORMAL) );
   }

   /* update all measurebuttons */
   ap_mbtn_update_all(wp->ud);
}

static void
ap_display_xcursor(WavePanel *wp, gint button, int x)
{
   UserData  *ud = wp->ud;
   AWCursor *csp;

   if ( button < 1 ) {
      return;
   }
   csp = ud->cursors[button - 1];
   if ( x == csp->x && csp->shown ) {
      return;
   }
   ap_update_xcursor(wp, csp, x);
}

/********************************************************************/

/*
 * color allocate
 */
static void
ap_style_find_color( GtkWidget *widget, gchar *name,
		     GdkColormap *colormap,  GdkColor *color)
{
   GtkStyle *style;

   if ( ! name ) {
      name = (gchar *) gtk_widget_get_name (widget);
   }
   style = gtk_widget_get_style (widget);
   if (  gtk_style_lookup_color ( style, name, color) ) {
      if ( ! gdk_colormap_alloc_color(colormap, color, FALSE, TRUE)) {
	 msg_fatal ( "gdk_color_alloc failed for %s", name );
      }
      msg_dbg ("%s: pixel 0x%x red %d, green %d, blue %d",
		 name, color->pixel, color->red, color->green, color->blue);
   } else {
      msg_fatal ("Can't find color for '%s'", name) ;
   }
}

void
ap_grid_color_allocate (WavePanel *wp, gint use_default)
{
   GdkColormap *colormap;

   colormap = gdk_window_get_colormap(wp->drawing->window);

   if ( use_default ) {
      ap_style_find_color( wp->drawing, "graticule", colormap, &wp->pg_color);
      wp->pg_gc =  gdk_gc_new(wp->drawing->window);
      if(! wp->pg_gc) {
	 msg_fatal ( "couldn't allocate graticule gc");
      }
   } else {
      if ( ! gdk_colormap_alloc_color(colormap, &wp->pg_color, FALSE, TRUE)) {
	 msg_fatal ("gdk_color_alloc failed");
      }
   }
   gdk_gc_set_foreground(wp->pg_gc, &wp->pg_color);
}

void
ap_wave_color_allocate (VisibleWave *vw, gint use_default)
{
   GdkColormap *colormap;

   colormap = gdk_window_get_colormap(vw->wp->drawing->window);

   if ( use_default ) {
      if ( ! vw->label) {
	 msg_warning( "(%s): label = NULL\n", vw->varname);
	 return;
      }
      ap_style_find_color( vw->label, NULL, colormap, &vw->color);
      vw->gc = gdk_gc_new(vw->wp->drawing->window);

      if(! vw->gc ) {
	 msg_fatal ( "couldn't allocate wave gc");
      }
   } else {
      if ( ! gdk_colormap_alloc_color(colormap, &vw->color, FALSE, TRUE)) {
	 msg_fatal ("gdk_color_alloc failed");
      }
      gtk_widget_modify_fg (vw->label, GTK_STATE_NORMAL, &vw->color);
      gtk_widget_modify_fg (vw->label, GTK_STATE_PRELIGHT, &vw->color);
      gtk_widget_modify_fg (vw->label, GTK_STATE_ACTIVE, &vw->color);
   }
   gdk_gc_set_foreground(vw->gc, &vw->color);
}

static void
ap_color_allocate (WavePanel *wp)
{
   UserData  *ud =   wp->ud;
   GdkColormap *colormap;
   GdkColor color;
   GdkColor bg_color;
   AWCursor *csp;
   gchar text[80];
   int i;

   colormap = gdk_window_get_colormap(wp->drawing->window);

   /* vertical bar cursors */
   for (i = 0 ; i < 2 ; i++) {
      csp = ud->cursors[i];

      ap_style_find_color( csp->label, csp->color_name, colormap, &color);
     
      /* we got the color, unshow the button */
      ap_widget_show(csp->button, FALSE);
      csp->cursor_gc = gdk_gc_new(wp->drawing->window);
      if ( ! csp->cursor_gc) {
	 msg_fatal ("couldn't allocate cursor %d gc", i);
      }
      /*
       * compute pixel to draw so XOR makes it come out right
       * on the background.
       */
      bg_color = wp->drawing->style->bg[GTK_WIDGET_STATE(wp->drawing)];
      color.pixel = color.pixel ^ bg_color.pixel;
      msg_dbg( "panel background pixel=0x%x rgb=%d,%d,%d",
	      bg_color.pixel,
	      bg_color.red,
	      bg_color.green,
	      bg_color.blue);
      msg_dbg( "panel %s pixel=0x%x rgb=%d,%d,%d",
	      csp->color_name,
	      color.pixel,
	      color.red,
	      color.green,
	      color.blue);
      gdk_gc_set_foreground(csp->cursor_gc, &color);
      gdk_gc_set_function(csp->cursor_gc, GDK_XOR);

   }
   csp = ud->cursors[2];
   /* unshow the 3rd button */
   ap_widget_show(csp->button, FALSE);

   if ( ! ud->hl_gc ){
      sprintf( text, "highlight");
      ap_style_find_color( wp->drawing, text, colormap, &color);
      ud->hl_gc =  gdk_gc_new(wp->drawing->window);
      if (! ud->hl_gc) {
	 msg_fatal ( "couldn't allocate %s gc", text);
      }
      gdk_gc_set_foreground(ud->hl_gc, &color);
   }
   if ( ! ud->srange->gc) {
      sprintf( text, "srange");
      ap_style_find_color( wp->drawing, text, colormap, &color);
      ud->srange->gc = gdk_gc_new(wp->drawing->window);
      if (! ud->srange->gc) {
	 msg_fatal ( "couldn't allocate %s gc", text);
      }
      gdk_gc_set_foreground(ud->srange->gc, &color);
      gdk_gc_set_background(ud->srange->gc, &bg_color);
      gdk_gc_set_function(ud->srange->gc, GDK_XOR);
   }
}
   
void
ap_draw_srange(SelRange *sr)
{
   g_assert(sr->gc != NULL);
   if (sr->type & SR_X) {
      gdk_draw_line(sr->wp->drawing->window, sr->gc,
		    sr->x1, sr->y1, sr->x2, sr->y1);
   }
   if (sr->type & SR_Y) {
      gdk_draw_line(sr->wp->drawing->window, sr->gc,
		    sr->x1, sr->y1, sr->x1, sr->y2);
   }
   if (sr->type == SR_XY) {
      gdk_draw_line(sr->wp->drawing->window, sr->gc,
		    sr->x1, sr->y2, sr->x2, sr->y2);
      gdk_draw_line(sr->wp->drawing->window, sr->gc,
		    sr->x2, sr->y1, sr->x2, sr->y2);
   }
}

void
ap_update_srange(SelRange *sr,  GdkEventMotion *event, int draw)
{
   int newx2, newy2;

   /*
    * the event->y does goofy things if the motion continues
    * outside the window, so we generate our own from the root
    * coordinates.
    */
   newx2 = event->x;
   newy2 = sr->y1 + (event->y_root - sr->y1_root);

   if (sr->drawn) {   /* undraw old */
      ap_draw_srange(sr);
   }
   sr->drawn = draw;
   if (sr->type & SR_X) {
      sr->x2 = newx2;
   }
   if (sr->type & SR_Y) {
      sr->y2 = newy2;
   }
   if (draw) {       /* draw new if requested */
      ap_draw_srange(sr);
   }
   msg_dbg( "type=%d newx=%d newy=%d draw=%d",
	   sr->type, sr->x2, sr->y2, draw);
   msg_dbg( "m %d %d %d %d",
	   (int) event->x, (int) event->y, 
	   (int) event->x_root, (int) event->y_root);
}

/*
 * done selecting range; do the callback
 */

void
ap_callback_srange(UserData *ud, WavePanel *wp )
{
   SelRange *sr = ud->srange;
   double xstart, ystart;
   double xend, yend;
   
   msg_dbg( "valid=%d type=%d x1=%d x2=%d  y1=%d y2=%d",
	   sr->wp->valid, sr->type, sr->x1, sr->x2, sr->y1, sr->y2);
   if ( ! sr->wp->valid ){
      return;
   }
   xstart = x2val(sr->wp, sr->x1, userPrefs->setLogX );
   xend   = x2val(sr->wp, sr->x2, userPrefs->setLogX );
      
   ystart = y2val(sr->wp, sr->y1 );
   yend   = y2val(sr->wp, sr->y2 );

   if ( ystart < yend ) {
      wp->start_yval = ystart;
      wp->end_yval = yend;
   } else {
      wp->start_yval = yend;
      wp->end_yval = ystart;
   }

   switch (sr->type) {
    case SR_X:
      am_cmd_zoom_absolute(ud, xstart, xend, userPrefs->setLogX );
      break;
    case SR_Y:
      wp->man_yzoom = 1;
      break;
    case SR_XY:
      wp->man_yzoom = 1;
      am_cmd_zoom_absolute(ud, xstart, xend, userPrefs->setLogX );
      break;
   }
}

/*
 * drawing area configure handler
 */
static gboolean
ap_configure_handler (GtkWidget *widget, GdkEventConfigure *event,
		      gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   int w = widget->allocation.width;
   int h = widget->allocation.height;
   static int colors_initialized = 0;
   
   msg_dbg( "w %d h = %d new %d %d",
	   w, h, event->width,  event->height);
   if ( wp->pixmap ){
      wp->width = w;
      wp->height = h;
      g_object_unref (wp->pixmap);
   }
   if ( ! colors_initialized ) {
      ap_color_allocate(wp);
      colors_initialized = 1;       
   }

   if ( ! wp->pg_gc ){
      ap_grid_color_allocate (wp, 1);
   }
   
   wp->pixmap = gdk_pixmap_new (widget->window,
                           widget->allocation.width,
                           widget->allocation.height, -1);

   /* Initialize the pixmap to bg color */
   gdk_draw_rectangle (wp->pixmap,
		       widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
		       TRUE,  0, 0, w, h );

   return TRUE;
}


/*
 *  draw one wave in a panel
 */
void
ap_draw_wave (VisibleWave *vw, WavePanel *wp)
{
   int x0, x1;
   int y0, y1;
   int i;
   double xstep;
   double xval;
   double yval;
   int w = wp->drawing->allocation.width;

   if ( ! vw->gc ) {
      ap_wave_color_allocate( vw, 1);      
   }
   g_assert(vw->gc != NULL);
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vw->button))) {
      gdk_gc_set_line_attributes(vw->gc, 2, GDK_LINE_SOLID,
				 GDK_CAP_BUTT, GDK_JOIN_ROUND);
   } else {
      gdk_gc_set_line_attributes(vw->gc, 1, GDK_LINE_SOLID,
				 GDK_CAP_BUTT, GDK_JOIN_ROUND);
   }

   xstep = (wp->end_xval - wp->start_xval) / w;  /* linear only */

   x1 = 0;
   yval = wv_interp_value(vw->var, wp->start_xval);
   y1 = val2y(wp, yval);

   for (i = 0, xval = wp->start_xval ; i < w ; i++ ) {
      x0 = x1; y0 = y1;
      x1 = x0 + 1;
      if (vw->var->wtable->iv->wds->min <= xval
	  && xval <= vw->var->wtable->iv->wds->max) {
	 yval = wv_interp_value(vw->var, xval);
	 y1 = val2y(wp, yval);
	 gdk_draw_line(wp->pixmap, vw->gc, x0, y0,  x1, y1);
      }
      if (userPrefs->setLogX) {
	 xval = x2val(wp, x0 + 1, userPrefs->setLogX );
      } else {
	 xval += xstep;
      }
   }
}

/*
 *  draw grid in a panel
 */
static void
ap_draw_grid (WavePanel *wp)
{
   guint i;
   guint step;
   guint segcnt = 0;
   GdkSegment *segp;
   GdkSegment seg[100];
   int w = wp->drawing->allocation.width;
   int h = wp->drawing->allocation.height;

   /* calculate grid segments */
   
   step = userPrefs->gridXSpacing ;
   msg_dbg( "xstep %d", step);
   for (segp = seg, i = 0; i < w ; i += step ) {
      segp->x1 = segp->x2 = i;
      segp->y1 = 0;
      segp->y2 = h - 1;
      segp++;
      segcnt++;
   }
   step = userPrefs->gridYSpacing ;
   msg_dbg( "ystep %d", step);
   for (i = 0 ; i < h ; i += step ) {
      segp->y1 = segp->y2 = i;
      segp->x1 = 0;
      segp->x2 = w - 1;
      segp++;
      segcnt++;
   }
   /* draw grid */
   gdk_draw_segments(wp->pixmap, wp->pg_gc, seg, segcnt);
}

/*
 * drawing area expose handler
 *  draw all the panel
 */
gboolean
ap_draw_wavepanel (GtkWidget *widget, GdkEventExpose *event, gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =   wp->ud;
   int w = widget->allocation.width;
   int h = widget->allocation.height;
   int i;
   int y;

   msg_dbg( "called");
   if ( ud->suppress_redraw ) {
      return FALSE;
   }

   /* clear all */
   gdk_draw_rectangle(wp->pixmap, 
		      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		      TRUE, 0, 0,  w, h);

   if (wp->selected) {
      gdk_draw_line(wp->pixmap, ud->hl_gc,     1, 1,      w - 2, 1);
      gdk_draw_line(wp->pixmap, ud->hl_gc, w - 2, 1,  w - 2, h - 2);
      gdk_draw_line(wp->pixmap, ud->hl_gc, w - 2, h - 2,  1, h - 2);
      gdk_draw_line(wp->pixmap, ud->hl_gc,     1, h - 2,  1,     1);
   }
   if ( wp->showGrid ) {
      /* graticule */
      ap_draw_grid (wp);
   } else {
      /* draw horizontal line at y=zero. */
      if (wp->start_yval < 0 && wp->end_yval > 0) {
	 y = val2y(wp, 0);
	 gdk_draw_line(wp->pixmap, wp->pg_gc, 0, y, w, y);
      }
   }
   /* draw waves */
   g_list_foreach(wp->vwlist, (GFunc) ap_draw_wave, wp); 

   /* draw the 2 cursors in the panel */
   for (i = 0 ; i < 2 ; i++) {                        
      AWCursor *csp = ud->cursors[i];
      if (csp->shown) {
	 gdk_draw_line(wp->pixmap, csp->cursor_gc, csp->x, 0, csp->x, h);
      }
   }

   /* draw select-range line, if in this WavePanel */
   if ( ud->srange && ud->srange->drawn && ud->srange->wp == wp) {
      ap_draw_srange(ud->srange);
   }
    
   if (event) {
      /* Draw the exposed portions of the pixmap in its window. */
      gdk_draw_drawable (widget->window,
			 widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			 wp->pixmap,
			 event->area.x, event->area.y,
			 event->area.x, event->area.y,
			 event->area.width, event->area.height);
   } else {
      /* Draw the whole thing. */
      gdk_draw_drawable (widget->window,
			 widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			 wp->pixmap,
			 0, 0, 0, 0, w, h);
   }
   return FALSE;
}

/*
 * drawing area button press handler
 */
static gboolean
ap_button_press_handler (GtkWidget *widget, GdkEventButton *event,
			 gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =   wp->ud;
   GdkCursor *cursor;
   
   msg_dbg( "button %d state %d mouseState %d",
	   event->button, event->state, ud->mouseState  );
   if ( wp->pixmap == NULL){
      return FALSE; /* we haven't gotten a configure event */
   }
   if ( event->button == 3 ){
      if ( wp->selected ) {
	 wp->selected = 0;
	 ud->selected_panel = NULL;
	 ap_panel_redraw(wp);
      }
      
      aw_update_toggle_state ( ud->ui,"/PanelPopup/PLogY", wp->logy );
      aw_update_toggle_state ( ud->ui,"/PanelPopup/PshowGrid", wp->showGrid );

      GtkWidget *menu = gtk_ui_manager_get_widget (ud->ui, "/PanelPopup");
      g_object_set_data (G_OBJECT(menu), "PanelPopup-action", (gpointer) wp);
      
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
		      NULL, NULL,  3, event->time);

      return TRUE;
   }
   switch (ud->mouseState) {
    case M_NONE:
      ud->selected_panel = wp;
      
      gtk_grab_add(widget);
      ud->mouseState = M_CURSOR_DRAG;
      ud->button_down = event->button;

      cursor = gdk_cursor_new(GDK_SB_H_DOUBLE_ARROW);
      gdk_window_set_cursor(widget->window, cursor);
      gdk_cursor_destroy(cursor);
      ud->drag_button = event->button;
      ap_display_xcursor(wp, ud->drag_button, event->x);
      ap_update_panel_selected( ud, event->state);
      wp->selected = 1;
      break;

    case M_SELRANGE_ARMED:
      gtk_grab_add(widget);
      ud->button_down = event->button;
      ud->mouseState = M_SELRANGE_ACTIVE;

      ud->srange->y1 = ud->srange->y2 = event->y;
      ud->srange->x1 = ud->srange->x2 = event->x;
      ud->srange->x1_root = event->x_root;
      ud->srange->y1_root = event->y_root;
      ud->srange->wp = wp;
      break;
      /* can't start another drag until first one done */

    case M_CURSOR_DRAG:
    case M_SELRANGE_ACTIVE:
    default:
      break;
   }
   ap_all_panel_redraw(ud);

   return TRUE;
}
/*
 * drawing area button release handler
 */
static gboolean
ap_button_release_handler (GtkWidget *widget, GdkEventButton *event,
			   gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =  wp->ud;
   
   msg_dbg( "button %d state %d", event->button, ud->mouseState );
   if ( wp->pixmap == NULL || ud->button_down != event->button ){
      return FALSE; /* we haven't gotten a configure event */
   }

   switch(ud->mouseState) {
    case M_CURSOR_DRAG:
      gtk_grab_remove(widget);
      gdk_window_set_cursor(widget->window, NULL);
      ap_display_xcursor(wp, ud->drag_button, event->x);
      ud->drag_button = -1;
      break;

    case M_SELRANGE_ACTIVE:
      gtk_grab_remove(widget);
      ap_set_all_wp_cursors(ud, -1);
      ap_update_srange(ud->srange, (GdkEventMotion *) event, 0);
      ap_callback_srange(ud, wp);
      break;

    default:
      break;
   }
   ud->mouseState = M_NONE;
   ap_panel_full_redraw(wp);
   return TRUE;

}
/*
 * drawing area motion handler
 */
static gboolean
ap_motion_handler (GtkWidget *widget, GdkEventMotion *event, gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =  wp->ud;
   
   msg_dbg( "state %d", ud->mouseState );
   if ( wp->pixmap == NULL){
      return FALSE; /* we haven't gotten a configure event */
   }

   switch(ud->mouseState) {
    case M_CURSOR_DRAG:
      ap_display_xcursor(wp, ud->drag_button, event->x);
      break;

    case M_SELRANGE_ACTIVE:
      ap_update_srange(ud->srange, event, 1);
      break;

    default:
      break;
   }

   return TRUE;
}

/*
 * drawing area dnd target drag_data_received handler
 *   Warning : key shift + dnd is action move by default.
 *   I set it to action copy,  so  dnd is action move.
 */
void 
ap_dnd_target_event (GtkWidget *widget, GdkDragContext *context,
		     gint  x, gint y,
		     GtkSelectionData   *data,
		     guint info, guint time, gpointer d)
{
   gboolean ret = TRUE;
   WavePanel *wp = (WavePanel *) d;
   DnDData *dd;
   gchar *s;
   gchar *name;
   
   msg_dbg( "info %d length %d", info, data->length );
   if ( info == TARGET_DVAR ) {
      gint dndMode = 1;
      dd = (DnDData *) data->data;
      
      if (data->length == sizeof(DnDData) &&
	  dd->magic == GAW_PRIVATE_DND_MAGIC) {
	 msg_dbg("  received dv = 0x%lx", (unsigned long) dd->dv);

	 if ( context->action == GDK_ACTION_MOVE ) {
	    msg_dbg("  action move received - Will copy ");
	    dndMode = 0;
	 }
	 if ( wp ) {
	    /* dest is drawing area */
	    msg_dbg(" adding visible wave to panel");
	    ap_var_to_panel(wp, dd->dv);
	 }

	 /* if wp == NULL, dest is toolbar delete wave button */
	 if ( dd->data && ( dndMode || wp == NULL ) ) {
	    /* move by default */
	    VisibleWave *vw = (VisibleWave *) dd->data ;
	    msg_dbg(" deleting visible wave");
	    ap_remove_wave_from_panel( vw->wp, vw);
	    ap_panel_full_redraw(vw->wp);
	 }
      } else {
	 msg_warning("bad magic number 0x%x", dd->magic);
      }
   } else  if ( info == TARGET_STRING ) {
      name = (gchar *) data->data;
      if ( ( s = strchr(name, '\r')) ) {
	 *s = 0;
      }
      msg_dbg( "  received '%s'", name );
      /* open the file as a data file */
	   
      gchar *str = "file://";
      if ( ( s = strstr(name, str)) ) {
	 s += strlen(str);
	 wp->ud->filename =  g_strdup(s);
	 ap_load_wave_file (wp->ud, NULL);
      }
    } else {
       ret = FALSE;
    }
  
  gtk_drag_finish (context, ret, FALSE, time);
}


/*
 * Set up widgets for a newly-created WavePanel -
 *    construct label and drawing areas
 */ 
void
ap_create_wave_panel(UserData *ud, WavePanel *wp)
{
   GtkWidget *drawing;
   
   msg_dbg("height = %d showXlabels = %d",
	  userPrefs->panelHeight, userPrefs->showXLabels);
   wp->start_xval = ud->start_xval;
   wp->end_xval = ud->end_xval;
   wp->req_height = userPrefs->panelHeight;
   wp->logy = userPrefs->setLogY;

   /* drawing area for waveform */
   drawing = gtk_drawing_area_new();
   wp->drawing = drawing;
   g_object_ref (drawing); /* increment ref to avoid destruction */

   g_signal_connect( drawing, "expose_event", 
		     G_CALLBACK (ap_draw_wavepanel), (gpointer) wp);
   g_signal_connect (drawing,"configure_event",
		     G_CALLBACK (ap_configure_handler), (gpointer) wp);

   g_signal_connect( drawing, "button_press_event", 
		     G_CALLBACK (ap_button_press_handler), (gpointer) wp);
   g_signal_connect( drawing, "button_release_event", 
		     G_CALLBACK (ap_button_release_handler), (gpointer) wp);
   g_signal_connect( drawing, "motion_notify_event", 
		     G_CALLBACK ( ap_motion_handler), (gpointer) wp);

   gtk_drawing_area_size(GTK_DRAWING_AREA(drawing), 
			 wp->width ? wp->width : userPrefs->panelWidth,
			 wp->req_height);
   gtk_widget_show(drawing);
   
   /* Set up a drawing as a drop target */
   ap_set_drag_dest(drawing, wp);

   gtk_widget_set_events(drawing, GDK_EXPOSURE_MASK |
			 GDK_BUTTON_RELEASE_MASK |
			 GDK_BUTTON_PRESS_MASK |
			 GDK_BUTTON1_MOTION_MASK |
			 GDK_BUTTON2_MOTION_MASK);
}

void
ap_create_label_measure_box(UserData *ud, WavePanel *wp)
{
   char lbuf[128];
   GtkWidget *vbox;
   GtkWidget *vbox1;
   GtkWidget *max_hbox;
   GtkWidget *min_hbox;
   GtkWidget *scrolled_window;

   /* the container : vbox */
   vbox = gtk_vbox_new(FALSE, 0);
   gtk_widget_show(vbox);
   wp->lmvbox = vbox;
   g_object_ref (vbox); /* increment ref to avoid destruction */

   /* hbox with logY and maxY labels */
   max_hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(vbox), max_hbox, FALSE, FALSE, 0);

   strcpy(lbuf, val2str(wp->end_yval, CNV_NORMAL));
   wp->label_max = gtk_label_new(lbuf);
   gtk_box_pack_end(GTK_BOX(max_hbox), wp->label_max,
		    FALSE, FALSE, 0);
   gtk_widget_show(wp->label_max);
 
   wp->label_ylog = gtk_label_new("LogY");
   gtk_box_pack_start(GTK_BOX(max_hbox), wp->label_ylog,
		      FALSE, FALSE, 0);
   ap_widget_show(wp->label_ylog, wp->logy);
  
   /*
    * table for buttons and masurements.
    * row 0 is space-filling dummy.
    */
   
   vbox1 = gtk_vbox_new(FALSE, 0);
   gtk_widget_show(vbox1);
   
   wp->lmtable = gtk_table_new(1, 3, FALSE); /* 1 row 3 col */
   gtk_widget_set_usize(wp->lmtable, userPrefs->lmtableWidth, -1);
   gtk_widget_show(wp->lmtable);

   gtk_table_attach(GTK_TABLE(wp->lmtable), vbox1,
		    /* X direction */       /* Y direction */
		    0, 3,                      0, 1,
		    0,
		    GTK_EXPAND|GTK_FILL,
		    0, 0 );

   /* wrap scrolled window with vertical scrollbar around table */
   scrolled_window = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW (scrolled_window),
                                  GTK_POLICY_NEVER, 
                                  GTK_POLICY_ALWAYS);
   gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW (scrolled_window),
				     GTK_CORNER_TOP_RIGHT);
   GTK_WIDGET_UNSET_FLAGS (GTK_SCROLLED_WINDOW (scrolled_window)->vscrollbar,
			   GTK_CAN_FOCUS);
   gtk_widget_show (scrolled_window);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
					 wp->lmtable);
   gtk_container_set_focus_vadjustment( GTK_CONTAINER (wp->lmtable),
      gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window)));

   gtk_box_pack_start(GTK_BOX(vbox), scrolled_window,
		      TRUE, TRUE, 0);

   /* last item in the outer vbox: hbox for minY label */
   min_hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_end(GTK_BOX(vbox), min_hbox,
		    FALSE, FALSE, 0);

   strcpy(lbuf, val2str(wp->start_yval, CNV_NORMAL));
   wp->label_min = gtk_label_new(lbuf);
   gtk_widget_show(wp->label_min);
   gtk_box_pack_end(GTK_BOX(min_hbox), wp->label_min,
		    FALSE, FALSE, 0);

   wp->min_hbox = min_hbox;
   wp->max_hbox = max_hbox;
   ap_widget_show(min_hbox, userPrefs->showYLabels);
   ap_widget_show(max_hbox, userPrefs->showYLabels);
}

/*
 * Delete a wavepanel line structure
 * and all data structures referenced from it.
 */

void
ap_delete_panel_line(UserData *ud, gint at_row)
{
   WavePanel *wp;
   WavePanel **owp;
   VisibleWave *vw;
   int i, j;

   if ( ud->npanels == 1 ) {
      msg_warning( "I don't want to remove the last panel!");
      return;
   }
   if ( at_row < 0 ) {
      at_row = ud->npanels - 1;
   }
   wp = ud->panels[at_row];
   while((vw = g_list_nth_data(wp->vwlist, 0)) != NULL) {
      ap_remove_wave_from_panel(wp, vw);
   }
   gtk_widget_destroy(wp->lmvbox);
   gtk_widget_destroy(wp->drawing);
   gdk_pixmap_unref(wp->pixmap);

   owp = ud->panels;
   ud->panels = g_new0(WavePanel*, ud->npanels - 1);

   ud->npanels--;
   for ( i = 0, j = 0 ; i < ud->npanels ; i++, j++ ){
      if ( j == at_row ) {
	 j++;
      }
      ud->panels[i] = owp[j];
   }
}

   
void
ap_create_panel_line(UserData *ud, gint at_row)
{
   WavePanel *wp;
   WavePanel **owp;
   int i, j;
   
   wp = g_new0(WavePanel, 1);
   wp->ud = ud;
   wp->valid = 1;
   wp->showGrid = userPrefs->showGrid;
   
   ap_create_label_measure_box(ud, wp);
   ap_create_wave_panel(ud, wp);

   owp = ud->panels;
   ud->panels = g_new0(WavePanel*, ud->npanels + 1);

   for ( i = 0, j = 0 ; j < ud->npanels ; i++, j++ ){
      if ( i == at_row ) {
	 ud->panels[i++] = wp;
      }
      ud->panels[i] = owp[j];
   }
   if ( at_row < 0 ){
      ud->panels[i] = wp;
   }
   ud->npanels++;
}


static void
ap_delete_panel_table(UserData *ud)
{
   GList *list = gtk_container_get_children(GTK_CONTAINER(ud->table));
   
   while (list) {
      GList *next = list->next;
      gtk_container_remove (GTK_CONTAINER(ud->table), list->data);
      list = next;
    }

   gtk_widget_destroy(ud->table);
   ud->table = NULL;
}

static void
ap_create_panel_table(UserData *ud )
{
   GtkWidget *table;
   int i;
   
   if (ud->table ) {
      /* delete the table */
      ap_delete_panel_table(ud);
   }
      
   /* create a gtk_table to store the WavePanels */
   table = gtk_table_new ( ud->npanels + 2, 2, FALSE); /* 4 row 2 columns */
   gtk_table_set_col_spacings(GTK_TABLE(table), 4);
   
   /* attach to table container */
   for (i = 0 ; i <  ud->npanels ; i++) {
      WavePanel *wp = ud->panels[i];       
      gtk_table_attach(GTK_TABLE(table), wp->lmvbox, 
		       /* X direction */       /* Y direction */
		       0, 1,                     i, i + 1, 
		       GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 1);
 
      gtk_table_attach(GTK_TABLE(table), wp->drawing, 
		       /* X direction */       /* Y direction */
		       1, 2,                     i, i + 1, 
		       GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 1);
   }

      
   /*  attach xlabel hbox to container */
   gtk_table_attach(GTK_TABLE(table), ud->xlabel_box, 
		    /* X direction */       /* Y direction */
		    1, 2,                    i, i + 1, 
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 1);
   i++;
   /*  attach status to container */
   gtk_table_attach(GTK_TABLE(table), ud->statusbar, 
		    /* X direction */       /* Y direction */
		    0, 1,                    i, i + 1, 
		    GTK_FILL, GTK_FILL, 0, 1);

   /*  attach scrollbar to container */
   gtk_table_attach(GTK_TABLE(table), ud->xscrollbar, 
		    /* X direction */       /* Y direction */
		    1, 2,                    i, i + 1, 
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 1);

   gtk_box_pack_start (GTK_BOX (ud->mainBox), table, TRUE, TRUE, 5);
   gtk_widget_show(table);
   ud->table = table;
}

gint
ap_find_panel_pos(UserData *ud, WavePanel *ppos)
{
   int i ;
   
   if ( ppos ) {
      for (i = 0 ; i < ud->npanels ; i++) {
	 if ( ud->panels[i] == ppos ){
	    return i;
	 }
      }
   }
   return -1;
}

WavePanel *
ap_find_panel_selected(UserData *ud)
{
   int i ;
   
   for (i = 0 ; i < ud->npanels ; i++) {
      if ( ud->panels[i]->selected ){
	 return ud->panels[i];
      }
   }
   return NULL;
}


/*
 * Remove this panel
 */
void
ap_remove_panel_line(UserData *ud, WavePanel *ppos)
{
   int pos = -1;
   
   if ( ! ppos ){
      if ( (ppos = ap_find_panel_selected(ud)) == NULL){
	 ppos = ud->panels[ud->npanels - 1];
      }
   }
   pos = ap_find_panel_pos(ppos->ud, ppos);

   ap_remove_all_wave_if_panel(ppos);

   ap_delete_panel_line(ppos->ud, pos); 
   ap_create_panel_table(ppos->ud);
   
   /* resize the window to minimum needed */
   gtk_window_resize (GTK_WINDOW (ud->window), userPrefs->min_win_width, userPrefs->min_win_height);
}
/*
 * Add a panel line, possibly at pos
 */

void
ap_add_panel_line(UserData *ud, WavePanel *ppos)
{
   int pos = ap_find_panel_pos(ud, ppos);
   
   ap_create_panel_line(ud, pos); 
   ap_create_panel_table(ud);
}

/*
 * Construct application widgets
 */

void ap_create_win_app (UserData *ud)
{
   int i;
  
   ud->srange = g_new0(SelRange, 1);
   
   /* create the 3 X measure buttons */
   ap_create_measure_unit(ud);

   for (i = 0 ; i < ud->reqpanels ; i++) {
      ap_create_panel_line(ud, -1);
   }
   /* horizontal box for X-axis labels */
   ap_create_xlabel_box(ud);
   ap_create_win_bottom(ud);
   
   /* create a container to store the WavePanels */
   ap_create_panel_table(ud); 
   
   /* status bar test */
   gtk_statusbar_push (GTK_STATUSBAR (ud->statusbar), 0, 
			  "Welcome to gaw");
}

/********************************************************************/

/*
 * drag and drop source callback
 */
void
ap_source_drag_data_get (GtkWidget *widget, GdkDragContext *context,
			 GtkSelectionData   *selection_data,
			 guint               info,
			 guint               time,
			 gpointer            data)
{
   WaveVar *dv = (WaveVar *) data;
   DnDData dd;
   
  if (info == TARGET_DVAR ) {
     dd.data = g_object_get_data (G_OBJECT(widget), "VW") ;

     dd.magic = GAW_PRIVATE_DND_MAGIC;
     dd.dv = dv;
     if ( prog_debug ){
	msg_dbg("dv = 0x%lx", (unsigned long) dv);
     }
     gtk_selection_data_set (selection_data,
			     selection_data->target,
			     8, (gpointer) &dd, sizeof(DnDData)); /* 8 = number of bits in a unit */
  } else if (info == TARGET_ROOTWIN) {
     msg_warning ("I was dropped on the rootwin");
  } else if (info == TARGET_STRING ) {
     msg_warning ("source type string");
  }
}

void  
ap_source_drag_data_delete  (GtkWidget *widget, GdkDragContext *context,
			     gpointer data)
{
  msg_dbg ("Delete the data!");
}
  
/*
 * initialize drag and drop in the list window
 */

void
ap_dnd_setup_source(DataFile *wdata, GtkWidget *button, WaveVar *wv, VisibleWave *vw)
{
   /* Drag site */
   gtk_drag_source_set (button, GDK_BUTTON1_MASK | GDK_BUTTON3_MASK,
		       target_table, n_targets, 
		       GDK_ACTION_COPY | GDK_ACTION_MOVE);
   gtk_drag_source_set_icon (button, 
			    gtk_widget_get_colormap (wdata->wlist_win),
			    wdata->drag_icon, wdata->drag_mask);

   g_object_set_data (G_OBJECT(button), "VW", vw);
   g_signal_connect (button, "drag_data_get",
		     G_CALLBACK (ap_source_drag_data_get), (gpointer) wv);
   g_signal_connect (button, "drag_data_delete",
		     G_CALLBACK (ap_source_drag_data_delete), (gpointer) wv );

}
/*
 * delete button, label, and measurement
 */
void
ap_delete_vis_wave_buttons(VisibleWave *vw, WavePanel *wp)
{
   int i;
   int row;
   UserData *ud = wp->ud;
   
   for ( i = 0 ; i < AW_NY_MBTN ; i++ ){
      gtk_widget_destroy(vw->mbtn[i]->button); /* kills the child label also */
      ud->all_measure_buttons = g_list_remove(ud->all_measure_buttons,
					      vw->mbtn[i]);
      g_free(vw->mbtn[i]);
   }
   row = am_gtk_table_get_child_row(wp->lmtable, vw->button);
   if (row == -1) {
      msg_warning( "VisibleWave not in this panel");
   } else {
      am_gtk_table_delete_row(wp->lmtable, row);
   }
   gtk_widget_destroy(vw->button);  /* kills the child label also */
}

gboolean
ap_vw_button_toggled_cb (GtkWidget *widget, gpointer data)
{
   ap_panel_redraw((WavePanel *) data);
   return FALSE;
}

gboolean
ap_vw_button_button_press_cb (GtkWidget *w,
			      GdkEventButton *event, gpointer data)
{
   VwWpInfo *vwwp =  (VwWpInfo *) data;
   WavePanel *wp = vwwp->wp;
   
//   gtk_widget_grab_focus (w);

   if (event->button == 3 && event->type == GDK_BUTTON_PRESS) {
      msg_dbg("called");
      GtkWidget *menu = gtk_ui_manager_get_widget (wp->ud->ui, "/ButtonPopup");
      g_object_set_data (G_OBJECT(menu), "ButtonPopup-action", data);
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
		      NULL, data, 3, event->time);
      return TRUE;
   }
   return FALSE;
}

/*
 * create button, label, and measurement
 * widgets for a new VisibleWave just added to a WavePanel
 */
void
ap_create_vis_wave_buttons(VisibleWave *vw, WavePanel *wp)
{
   DataFile *adf = vw->adf;
   UserData *ud = wp->ud;
   char lbuf[64];
   char temp[64];
   char *p = "";
   int newrow;
   int n, l;
   int i;

   g_assert(adf != NULL);

   if ( vw->var->wtable->wf->tables->len > 1 ) {
      /* wv_is_multisweep */
      sprintf( temp, "  @ %.10s=%g", 
	       vw->var->wtable->name, vw->var->wtable->swval);
      p = temp;
   }
   l = sizeof(lbuf) - 2 - 10 - strlen(p);
   n = MIN(l, 15);
   snprintf(lbuf, sizeof(lbuf), "%d: %.*s%s", adf->ftag, n, vw->varname, p);
   
   p = "Visible Wave button:\n\nClick button 1\n to select wave\n\nPress button 3\n for options menus\n\nDrag button to move,\n copy or delete wave";
   vw->button = ap_create_toggle_button(NULL, "wavebutton", 
				     p, ud->tooltips );
   g_signal_connect (vw->button, "toggled",
		     G_CALLBACK (ap_vw_button_toggled_cb), (gpointer) wp );
   g_signal_connect (vw->button, "button_press_event",
		     G_CALLBACK (ap_vw_button_button_press_cb),
		     (gpointer) vw->vwwp);

   ap_dnd_setup_source(vw->adf, vw->button, vw->var, vw);

   sprintf(temp, "wavecolor%d", vw->colorn);
   vw->label = ap_create_measure_label(vw->button, lbuf, temp ); 

   /*
    * create new row #1 of the label/measurement table,
    * and add this stuff there.  row 0 reserved for Y-axis label.
    */
   newrow = 1;
   am_gtk_table_insert_row(wp->lmtable, newrow);
   gtk_table_attach(GTK_TABLE(wp->lmtable), vw->button, 
		    0, 1, newrow, newrow + 1,
		    GTK_EXPAND|GTK_FILL,
		    0,
		    0, 0 );

   /* create Y measurement buttons */
   for ( i = 0 ; i < AW_NY_MBTN ; i++ ) {
      GtkWidget *button;
      
      vw->mbtn[i] = g_new0(MeasureBtn, 1);
      vw->mbtn[i]->var = vw->var;
      ud->all_measure_buttons = g_list_prepend(ud->all_measure_buttons,
					       vw->mbtn[i]);

      if ( i == 2 ){
	 sprintf(lbuf, "cursorD");
	 sprintf(temp, "measure: cursor1 - cursor0");
      } else {
	 sprintf(lbuf, "cursor%d", i);
	 sprintf(temp, "measure: variable(cursor%d)", i);
      }
      button = ap_create_measure_button( NULL, "wavebutton", 
			temp, ud->tooltips );
      vw->mbtn[i]->label = ap_create_measure_label(button, "0.0", lbuf ); 
      gtk_widget_set_usize(button, 60, -1);
     
      vw->mbtn[i]->button = button;
      gtk_table_attach(GTK_TABLE(wp->lmtable), button,
		       i + 1, i + 2,             newrow, newrow + 1,
		       GTK_FILL,
		       0,
		       0, 0 );
      
      vw->mbtn[i]->csp = ud->cursors[i];
      ap_mbtn_update(vw->mbtn[i], ud);
   }
}

/*
 * remove waveform from panel
 *      Somthing bad will probably happen if the waveform isn't actually
 *      in the indicated panel.
 */
void
ap_remove_wave_from_panel(WavePanel *wp, VisibleWave *vw)
{
   if ( wp->vwlist ){
      wp->vwlist = g_list_remove(wp->vwlist, vw);
   }

   ap_delete_vis_wave_buttons(vw, wp);

   gdk_gc_destroy(vw->gc);
   g_free(vw->vwwp);
   g_free(vw->varname);

   g_free(vw);

}
  
/*
 * Add a new waveform to a WavePanel, creating a new VisibleWave.
 * If no wavepanel is specified, try to use the first "selected" wavepanel,
 * This is the only place that VisibleWave structures are created.
 */

void
ap_var_to_panel(WavePanel *wp, WaveVar *dv)
{
   VisibleWave *vw;
   UserData *ud = userData;
        
   if (wp == NULL) {
      if ( (wp = ud->selected_panel)  == NULL) {
	 msg_dbg("no default found");
	 return;
      }
   }

   vw = g_new0(VisibleWave, 1);
   vw->wp = wp;
   vw->var = dv;
   vw->varname = g_strdup(dv->sv->name);
   vw->adf = dv->wtable->wf->udata;

   vw->colorn = wp->nextcolor;
   wp->nextcolor = (wp->nextcolor + 1) % ud->NWColors;

   wp->vwlist = g_list_append(wp->vwlist, vw);
   vw->valid = 1;
   vw->vwwp = g_new0(VwWpInfo, 1);
   vw->vwwp->wp = wp;
   vw->vwwp->vw = vw;

   /* add button to Y-label box */
   if (wp->lmtable) {
      ap_create_vis_wave_buttons(vw, wp);
   }
   /* update max and min in wp and ud */
   ap_panel_full_redraw(wp);

   /* Update y-axis labels */
   ap_draw_panel_ylabels(wp);

   return;
}

/*
 * Called for all button presses on wavelist button.
 * If it is a doubleclick, add variable to the "current" wavepanel immediately.
 */
static gint
ap_list_button_cb(GtkWidget *widget, GdkEventButton *bevent, gpointer data)
{
   WaveVar *dv = (WaveVar *) data;
   if (bevent->type == GDK_2BUTTON_PRESS) {
      msg_dbg( "doubleclicked %s", dv->sv->name);
      ap_var_to_panel(NULL, dv);
   }
   /*
    * TRUE to stop other handlers from being invoked for the event.
    *  FALSE to propagate the event further.
    */
   return TRUE;
}

/*
 * Add a button for each variable in the file to the win_wlist box for it.
 * Arrange for the buttons to be drag-and-drop sources for placing the
 * variables into wavepanels.
 */

void
ap_add_list_button(gpointer d, gpointer p)
{
   WaveVar *wv = (WaveVar *) d;
   DataFile *wdata = (DataFile *) p;
   GtkWidget *button;
   char lab[128];

   sprintf(lab, "%s", wv->sv->name); 
   if( wv->wtable->wf->tables->len > 1) {
      /* wv_is_multisweep */
      char lab[4096];
      sprintf(lab, "%s @ %s=%g", wv->sv->name, 
	      wv->wtable->name, wv->wtable->swval);
      button = gtk_button_new_with_label(lab);
   }
                
   button = gtk_button_new_with_label(lab);
   gtk_widget_set_name(button, "listButton" );
   gtk_box_pack_start (GTK_BOX (wdata->wlist_box), button, FALSE, FALSE, 0);
   gtk_widget_show (button);
   gtk_tooltips_set_tip(GTK_TOOLTIPS(wdata->ud->tooltips), button,
			"Wavefile Variable.\nDrag-and-Drop to a WavePanel.", "");

   ap_dnd_setup_source(wdata, button, wv, NULL);

   g_signal_connect (button, "button-press-event",
		     G_CALLBACK (ap_list_button_cb ), 
		     (gpointer) wv);

}

/*
 * Show the variable-list window for a waveform data file.
 * If the window already exists, simply raise it to the top.
 */
void
ap_create_list_win (UserData *ud,  DataFile *wdata)
{
   GtkWidget *box1;
   GtkWidget *scrolled_window;
   GtkWidget *label;
   char buf[256];
   int x, y;

   if ( ! wdata) {
      msg_warning("wdata is NULL");
      return;
   }
   if( wdata->wlist_win) {
      gdk_window_raise(wdata->wlist_win->window);
      return;
   }
   
   wdata->wlist_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_widget_set_name(wdata->wlist_win, "data_list");
   sprintf(buf, "Gaw: %.64s", wdata->filename);
   gtk_window_set_title(GTK_WINDOW(wdata->wlist_win), buf);
   gtk_widget_set_usize(wdata->wlist_win, 150, 300);
   
   /* this set set wlist_win = NULL at destroy event */
   g_signal_connect (wdata->wlist_win, "destroy",
		     G_CALLBACK (gtk_widget_destroyed),
		     &(wdata->wlist_win) );


   gtk_window_get_position(GTK_WINDOW(ud->window), &x, &y);

   /* a vertical box */
   box1 = gtk_vbox_new(FALSE, 0);
   gtk_container_add(GTK_CONTAINER(wdata->wlist_win), box1);
   gtk_widget_show(box1);
   wdata->wlist_vbox = box1;
   
   /* create the menu */
   aw_vl_create_ui_manager ( wdata );
      
   /* a label */
   sprintf(buf, "%d: %.64s", wdata->ftag, wdata->label);
   label = gtk_label_new(buf);
   gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
   gtk_widget_show(label);
   gtk_box_pack_start (GTK_BOX (box1), label, FALSE, FALSE, 0);

   scrolled_window = gtk_scrolled_window_new (NULL, NULL);
   gtk_container_border_width (GTK_CONTAINER (scrolled_window), 10);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				   GTK_POLICY_AUTOMATIC, 
				   GTK_POLICY_ALWAYS);
   GTK_WIDGET_UNSET_FLAGS (GTK_SCROLLED_WINDOW (scrolled_window)->vscrollbar,
			   GTK_CAN_FOCUS);
   gtk_box_pack_start(GTK_BOX (box1), scrolled_window, TRUE, TRUE, 0);
   gtk_widget_show (scrolled_window);

   wdata->wlist_box = gtk_vbox_new (FALSE, 0);
   gtk_container_border_width (GTK_CONTAINER (wdata->wlist_box), 10);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
					 wdata->wlist_box);

   gtk_container_set_focus_vadjustment( GTK_CONTAINER (wdata->wlist_box),
					gtk_scrolled_window_get_vadjustment(
					GTK_SCROLLED_WINDOW(scrolled_window)));
   gtk_widget_show (wdata->wlist_box);
   
   wdata->drag_icon = gdk_pixmap_colormap_create_from_xpm_d (NULL,
		     gtk_widget_get_colormap (wdata->wlist_win),
		     &wdata->drag_mask,
		     NULL, wave_drag_ok_xpm);
   
   wf_foreach_wavevar(wdata->wf, ap_add_list_button, (gpointer) wdata);

   gtk_widget_show(wdata->wlist_win);
}
 
DataFile *
ap_load_wave_file (UserData *ud, DataFile *wdata)
{
   static int next_tag = 0;

   if ( ! wdata ) {
      wdata = g_new0(DataFile, 1);
      wdata->filename = ud->filename;
      wdata->ftag = next_tag++;
      wdata->ud = ud;

      char *cp = wdata->filename ;
      if (strlen(wdata->filename) > 16) {
	 cp = strrchr(wdata->filename, '/');
	 if (cp) {
	    cp++;
	 } else {
	    cp = wdata->filename;
	 }
      }
      wdata->label = g_strdup(cp);
      
      ud->wdata_list = g_list_prepend(ud->wdata_list, wdata);
      aw_vl_menu_item_add( ud, wdata);
   }

   wdata->wf = wf_read(wdata->filename, 0);

   if (wdata->wf == NULL) {
      ap_delete_wave_file( NULL, wdata);
      return NULL;
   }

   wdata->wf->udata = wdata;

   if ( ! wdata->wlist_win ) {
      ap_create_list_win (ud, wdata);
   }

   return wdata;
}

/*
 * remove_wfile_waves - 
 * Remove from their respective panels all waveforms from the specified file.
 * This works just like delete_selected_waves, except for the test
 * that puts them on the to-delete list.  
 * TODO: factor out this common code.
 */

typedef struct _TestParams {
   VisibleWave *vw;
   WavePanel *wp;
   gpointer *data;
} TestParams ;

/*
 * return TRUE if visible wave is selected
 */
static gboolean
ap_wave_if_selected(TestParams *params)
{
   return gtk_toggle_button_get_active
      (GTK_TOGGLE_BUTTON(params->vw->button)) ;
}

/*
 * return TRUE if visible wave  reference this file
 */
static gboolean
ap_wave_if_wfile(TestParams *params)
{
   if ( params->vw->adf == (DataFile *) params->data) {
      return TRUE;
   }
   return FALSE;
}

/*
 * return TRUE if visible wave  reference this panel
 */
static gboolean
ap_wave_if_panel(TestParams *params)
{
   if ( params->wp == (WavePanel *) params->data) {
      return TRUE;
   }
   return FALSE;
}

/*
 * Remove from panels any VisibleWaves that responds TRUE
 * to the specified test.
 */

static void
ap_remove_all_wave_if_test(UserData *ud,TestParams *params,
			    gboolean  (*testFunc) () )
{
   int i;
   GList *oldlist;
   GList *newlist;
   GList *list;
   WavePanel *wp;
   VisibleWave *vw;

   for (i = 0 ; i < ud->npanels; i++) {
      wp = ud->panels[i];
      params->wp = wp;
      list = oldlist = wp->vwlist;
      newlist = wp->vwlist = NULL;

      while (list) {
	 GList *next = list->next;
	 
	 vw = params->vw = (VisibleWave *) list->data;
	 if ( testFunc ( params ) ) {
	    ap_remove_wave_from_panel(wp, vw);
	 } else {
	    newlist = g_list_prepend(newlist, vw);
	 }
	 list = next;
      }
      wp->vwlist = newlist;
      g_list_free(oldlist);
      
      ap_panel_full_redraw(wp);
   }
}

/*
 * Remove from panels any VisibleWaves that have been
 * selected by clicking on their label-buttons.
 */
void
ap_remove_all_wave_selected(UserData *ud)
{
   TestParams *params; 
   params = g_new0(TestParams, 1);

   ap_remove_all_wave_if_test(ud, params, ap_wave_if_selected );
   g_free(params);
}

/*
 * Delete a wave file.
 * callback from menu: wavelist->file->delete
 */

void
ap_remove_all_wave_if_wfile(DataFile *wdata)
{
   TestParams *params; 
   UserData *ud = wdata->ud;

   params = g_new0(TestParams, 1);
   params->data = (gpointer) wdata;
   ap_remove_all_wave_if_test(ud, params, ap_wave_if_wfile );
   g_free(params);
}

/*
 * Delete all waves from panel.
 */

void
ap_remove_all_wave_if_panel(WavePanel *wp)
{
   TestParams *params; 
   UserData *ud = wp->ud;

   params = g_new0(TestParams, 1);
   params->data = (gpointer) wp;
   ap_remove_all_wave_if_test(ud, params, ap_wave_if_panel );
   g_free(params);
}

/*
 * Repoint all visible waves that reference this file to the new data.   
 * Any visiblewaves referencing variables that no longer exist in the new file
 * are deleted.
 */
void
ap_update_wfile_waves( DataFile *wdata)
{
   UserData *ud = wdata->ud;
   int i;
   int j;
   GList *oldlist;
   GList *newlist;
   GList *list;
   WavePanel *wp;
   VisibleWave *vw;
   WaveVar  *wv;

   for (i = 0 ; i < ud->npanels; i++) {
      wp = ud->panels[i];
      list = oldlist = wp->vwlist;
      newlist = wp->vwlist = NULL;

      while (list) {
	 GList *next = list->next;
	 
	 vw = (VisibleWave *) list->data;
	 if (vw->adf == wdata) {
	    wv = wf_find_variable(wdata->wf, vw->var->sv->name,
				  vw->var->wtable->swindex );
	    if (wv) {
	       msg_dbg( "updated variable %s to %lx",
		       vw->varname, (unsigned long) wdata->wf);
	       vw->var = wv;
	       for (j = 0 ; j < AW_NY_MBTN ; j++) {
		  vw->mbtn[j]->var = wv;
	       }
	       newlist = g_list_prepend(newlist, vw);
	    } else {
	       msg_dbg( "variable %s no longer in file %s; removing from panel",
		       vw->varname, wdata->wf->ss->filename);
	       ap_remove_wave_from_panel(wp, vw);
	    }
	 } else {
	    newlist = g_list_prepend(newlist, vw);
	 }
	 list = next;
      }
      wp->vwlist = newlist;
      g_list_free(oldlist);
      
      ap_panel_full_redraw(wp);
   }
}

/*
 * command or callback from menu: wavelist->file->reload
 */
void
ap_reload_wave_file(DataFile *wdata)
{
   WaveFile *old_wf;
   UserData *ud = wdata->ud;

   old_wf = wdata->wf;
   
   wdata = ap_load_wave_file (ud, wdata);
   if ( wdata->wf == NULL) {
      return;
   }
   msg_dbg( "(%s) old=%lx new=%lx",
	   wdata->wf->ss->filename,
	   (unsigned long) old_wf, (unsigned long) wdata->wf );

   ap_update_wfile_waves(wdata);
   
   /* remove old buttons from list, and add new ones */    
   if (wdata->wlist_win && GTK_WIDGET_VISIBLE(wdata->wlist_win)) {
      gtk_container_foreach(GTK_CONTAINER(wdata->wlist_box),
			    (GtkCallback) gtk_widget_destroy, NULL);
      wf_foreach_wavevar(wdata->wf, ap_add_list_button, (gpointer) wdata);
   }

   wf_free(old_wf);
   ap_mbtn_update_all(ud);
}


/*
 * Delete a wave file.
 * callback from menu: wavelist->file->delete
 */
void
ap_delete_wave_file(GtkWidget *w, DataFile *wdata)
{
   UserData *ud = wdata->ud;
   
   if ( wdata->wf ) {
      /* remove references from displayed waves */
      ap_remove_all_wave_if_wfile(wdata);

      /* remove per-file GUI stuff */
      if (wdata->wlist_win && GTK_WIDGET_VISIBLE(wdata->wlist_win)) {
	 gtk_widget_destroy(wdata->wlist_win);
      }
	
      /* now nuke the data */
      wf_free(wdata->wf);
      wdata->wf = NULL;
   }
   g_free(wdata->label);
   ud->wdata_list = g_list_remove(ud->wdata_list, wdata);
   
   msg_dbg( "free DataFile 0x%lx", (unsigned long) wdata);
   if ( wdata->drag_icon ) {
      g_object_unref (wdata->drag_icon);
      g_object_unref (wdata->drag_mask);
   }
   g_free(wdata);
}
