/*
 * gaw - Gtk analog waveform viewer
 *   GTK interface to Gaw
 * 
 * include LICENSE
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <libgen.h>
#include <stdarg.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#include <gaw.h>

/*
 * hq Analog wave globals
 */
char *progname ;
UserData  *userData  = NULL;

/*
 * Global functions
 */
void aw_do_exit (GtkWidget *widget, GtkWidget *window);
GtkWidget *do_editable_cells (GtkWidget *do_widget);

/*
 * save configuration file
 */
void aw_save_config_action (GtkAction *action, UserData *ud )
{
   userPrefs->npanels = ud->npanels;

   rc_save_config(ud->prog, ud->prog);
}

/*
 * get data for a menu
 */
static gpointer
aw_get_menu_data (GtkUIManager *ui, gchar *menuName,  gchar *actionName)
{
   msg_dbg("aw_get_menu_data menu: %s %s", menuName, actionName );
   gpointer data;
   GtkWidget *menu = gtk_ui_manager_get_widget (ui, menuName);

   if ( (data = g_object_get_data (G_OBJECT(menu), actionName))){
      msg_dbg("%s %s received 0x%lx", menuName, actionName, (unsigned long) data );
      g_object_set_data (G_OBJECT(menu), actionName, NULL);
   }
   return data;
}


static gint
aw_color_dialog (GtkWidget *window, GdkColor *color, gchar *title )
{
   GtkWidget *dialog;
   GtkColorSelection *colorsel;
   gint response;

   dialog = gtk_color_selection_dialog_new (title);
   
   gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  
   colorsel = GTK_COLOR_SELECTION (GTK_COLOR_SELECTION_DIALOG (dialog)->colorsel);

   gtk_color_selection_set_previous_color (colorsel, color);
   gtk_color_selection_set_current_color (colorsel, color);
  
   response = gtk_dialog_run (GTK_DIALOG (dialog));
   gtk_color_selection_get_current_color (colorsel, color);

   gtk_widget_destroy (dialog);

   return response;
}

static void
aw_wave_color_action (GtkAction *action, UserData *ud )
{
   gint response;
   VwWpInfo *vwwp =  (VwWpInfo *) aw_get_menu_data (ud->ui, "/ButtonPopup",
						    "ButtonPopup-action" );

   response = aw_color_dialog( ud->window,  &vwwp->vw->color, "Changing wave color");
  
   if (response == GTK_RESPONSE_OK) {
      ap_wave_color_allocate (vwwp->vw, 0);
  
      ap_panel_redraw(vwwp->wp);
   }
}

static void
aw_grid_color_action (GtkAction *action, UserData *ud )
{
   gint response;
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );

   if ( ! wp ) {
      return;
   }
   response = aw_color_dialog( ud->window, &wp->pg_color, "Changing grid color");
   
   if (response == GTK_RESPONSE_OK) {
      ap_grid_color_allocate (wp, 0);

      ap_panel_redraw(wp);
   }
}


void aw_add_if_writable (GdkPixbufFormat *data, GSList **list)
{
   if (gdk_pixbuf_format_is_writable (data)) {
      *list = g_slist_prepend (*list, data);
   }
}

typedef struct _DialogData {
   GtkWidget *dialog;
   UserData *ud;
   WavePanel *wp;
} DialogData;

void
aw_fs_radio_clicked (GtkWidget *w,  DialogData *pdata )
{
   char *save_filename;
   UserData *ud =  pdata->ud;
   
   ud->imgFmt = gtk_button_get_label (GTK_BUTTON (w));
   save_filename = g_strdup_printf( "%s%02d.%s", userPrefs->imgBaseName,
			    ap_find_panel_pos( ud, pdata->wp), ud->imgFmt);       
   gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (pdata->dialog), save_filename);
//   fprintf( stderr, "2 format : %s\n", userData->imgFmt );
   g_free(save_filename);
}

GtkWidget *
aw_create_radio_widget( DialogData *pdata )
{
   UserData *ud =  userData;
   GSList *group = NULL;
   GtkWidget *button;
   GtkWidget *hbox;
   char *name;
   
   GSList *formats = gdk_pixbuf_get_formats ();
   if ( ud->imgFormats == NULL ) {
      g_slist_foreach (formats, (GFunc) aw_add_if_writable,  &ud->imgFormats);
   }
   g_slist_free (formats);

   hbox = gtk_hbox_new (FALSE, 5);

   GSList *list =  ud->imgFormats;   
   while (list) {
      GSList *next = list->next;
 
      name = gdk_pixbuf_format_get_name  (list->data);
      button = gtk_radio_button_new_with_label (group, name);
      group = gtk_radio_button_group (GTK_RADIO_BUTTON (button));
      g_signal_connect (button, "clicked",
			(GtkSignalFunc) aw_fs_radio_clicked, pdata );
      gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
      gtk_widget_show (button);
      if ( strcmp( name, userPrefs->imgFmt) == 0 ){
	 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
	 ud->imgFmt = name;
      }
      list = next;
    }
   gtk_widget_show (hbox);
   return hbox;
}



static void
aw_save_img_file_dialog ( WavePanel *wp, GdkPixbuf *pixbuf)
{
   DialogData *pdata = NULL;
   UserData *ud = wp->ud; 
   GtkWidget *dialog;
   GtkWidget *hbox;
   gint response;
   char *save_filename;
   GError *err = NULL;

   if (! pdata) {
      pdata = g_new0 (DialogData, 1);
   }

   dialog = gtk_file_chooser_dialog_new ("Select file",
					 GTK_WINDOW (userData->window),
					 GTK_FILE_CHOOSER_ACTION_SAVE,
					 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					 GTK_STOCK_SAVE, GTK_RESPONSE_OK,
					 NULL);
   pdata->dialog = dialog;
   pdata->ud = ud;
   pdata->wp = wp;
   hbox = aw_create_radio_widget(pdata);
   gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER (dialog), hbox);
   
   gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
   response = gtk_dialog_run (GTK_DIALOG (dialog));

   if (response == GTK_RESPONSE_OK) {
      int ret;
      gchar tmpstr[32];
      save_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      if ( strcmp( ud->imgFmt, "jpeg") == 0 ){
	 sprintf(tmpstr, "%d",  userPrefs->jpegQuality );
	 ret = gdk_pixbuf_save (pixbuf, save_filename, ud->imgFmt, &err,
                                      "quality", tmpstr,  NULL) ;
      } else if ( strcmp( ud->imgFmt, "png") == 0 ){
	 sprintf(tmpstr, "%d",  userPrefs->pngCompression );
	 ret = gdk_pixbuf_save (pixbuf, save_filename, ud->imgFmt, &err,
                                      "compression", tmpstr,  NULL) ;
      } else {
	  ret = gdk_pixbuf_save (pixbuf, save_filename, ud->imgFmt, &err,
                                      NULL) ;
      }
      if ( ! ret ) {
	 msg_warning ("%s", err->message);
	 g_error_free (err);
      }
   }
   
   gtk_widget_destroy (dialog);
   g_free(pdata);
}

static void
aw_export_to_img_action (GtkAction *action, UserData *ud )
{
   GdkPixbuf     *pixbuf;

   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );
   if ( wp ) {
      ap_draw_wavepanel(wp->drawing, NULL, (gpointer) wp);
//      getchar();
      pixbuf = gdk_pixbuf_get_from_drawable (NULL, wp->drawing->window, NULL,
                                               0, 0, 0, 0,
					     wp->drawing->allocation.width,
					     wp->drawing->allocation.height);
      aw_save_img_file_dialog (wp, pixbuf);
      g_object_unref (pixbuf);
   }
}

static void
toggle_bar_style_action (GtkAction *action, UserData *ud )
{
   GtkToolbar *tb =  GTK_TOOLBAR(ud->toolBar);
   GtkToolbarStyle style = gtk_toolbar_get_style(tb);

   if ( style == GTK_TOOLBAR_TEXT ){
      gtk_toolbar_set_style(tb, GTK_TOOLBAR_ICONS );
   } else {
      gtk_toolbar_set_style(tb, GTK_TOOLBAR_TEXT );
   }
}

static void
aw_open_file_action (GtkAction *action,  UserData *ud)
{
   GtkWidget *dialog;
   gint response;
   char *open_filename;
  
   dialog = gtk_file_chooser_dialog_new ("Select file",
					 GTK_WINDOW (userData->window),
					 GTK_FILE_CHOOSER_ACTION_OPEN,
					 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					 GTK_STOCK_OPEN, GTK_RESPONSE_OK,
					 NULL);
   gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
   response = gtk_dialog_run (GTK_DIALOG (dialog));

   if (response == GTK_RESPONSE_OK) {
      open_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      if ( ud->filename ) {
	 g_free (ud->filename );
      }
      ud->filename = open_filename;
      ap_load_wave_file (ud, NULL);
   }
   
   gtk_widget_destroy (dialog);
}


static void
activate_action (GtkAction *action, UserData *ud )
{
   const gchar *name = gtk_action_get_name (action);

   msg_dbg ("Action \"%s\" activated", name );
   if ( strcmp( name, "Quit") == 0 ){
      aw_do_exit(ud->window, ud->window);
   }
}

static void
aw_redraw_all_action (GtkAction *action, UserData *ud )
{
   int i;
   
   for (i = 0 ; i < ud->npanels; i++) {
      WavePanel *wp = ud->panels[i];
      
      ap_panel_full_redraw(wp);
   }
}

static void
aw_showXlabel_action (GtkToggleAction *action, UserData *ud )
{
   userPrefs->showXLabels = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   msg_dbg ("Action showXlabel %d", userPrefs->showXLabels );
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   ap_widget_show(ud->xlabel_box, userPrefs->showXLabels);
}

static void
aw_showYlabel_action (GtkToggleAction *action, UserData *ud )
{
   int i;
   
   userPrefs->showYLabels = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   msg_dbg ("Action showYlabel %d", userPrefs->showYLabels );
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   for (i = 0 ; i < ud->npanels; i++) {
      WavePanel *wp = ud->panels[i];
      
      ap_widget_show(wp->min_hbox, userPrefs->showYLabels);
      ap_widget_show(wp->max_hbox, userPrefs->showYLabels);
   }
}

static void
aw_logx_action (GtkToggleAction *action, UserData *ud )
{
   userPrefs->setLogX = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   aw_redraw_all_action ((GtkAction *) action, ud);
}

static void
aw_ydiff_action (GtkToggleAction *action, UserData *ud )
{
   userPrefs->showYDiff = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   ap_mbtn_update_all(ud);
}

static void
aw_logy_action (GtkToggleAction *action, UserData *ud )
{
   int i;
   
   userPrefs->setLogY = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   
   for (i = 0 ; i < ud->npanels; i++) {
      WavePanel *wp = ud->panels[i];
      
      wp->logy =  userPrefs->setLogY;
      ap_widget_show(wp->label_ylog, wp->logy);
      ap_panel_full_redraw(wp);
   }
}

static void
aw_logy_pop_action (GtkAction *action, UserData *ud )
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );
   msg_dbg("aw_logy_pop_action" );
   if ( wp ) {
      wp->logy = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
      ap_widget_show(wp->label_ylog, wp->logy);
      ap_panel_full_redraw(wp);
   }
}


static void
aw_show_grid_action (GtkToggleAction *action, UserData *ud )
{
   int i;
   
   userPrefs->showGrid = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
   if ( ud->xlabel_box == NULL ){ 
      return; /* widget not yet created */
   }
   
   for (i = 0 ; i < ud->npanels; i++) {
      WavePanel *wp = ud->panels[i];
      
      wp->showGrid =  userPrefs->showGrid;
      ap_panel_full_redraw(wp);
   }
}

static void
aw_show_grid_pop_action (GtkAction *action, UserData *ud )
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );
   if ( wp ) {
      msg_dbg("showGrid = %d", wp->showGrid );
      wp->showGrid = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
      msg_dbg("showGrid = %d", wp->showGrid );
      ap_panel_full_redraw(wp);
   }
}


static void
aw_add_panel_above_action (GtkAction *action, UserData *ud )
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );
   
   if ( wp ) {
      ap_add_panel_line(userData, wp);
   }
}

static void
aw_add_panel_action (GtkAction *action,  UserData *ud)
{
   ap_add_panel_line(ud, 0);
}

static void
aw_remove_panel_action (GtkAction *action,  UserData *ud)
{
   ap_remove_panel_line(ud, 0);
}

static void
aw_remove_this_panel_action (GtkAction *action, UserData *ud )
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );

   if ( wp ) {
      ap_remove_panel_line(userData, wp);
   }
}


static void
aw_zoom_in_action (GtkAction *action, UserData *ud )
{
   double start;
   double end;
   
   start = ud->start_xval + ( ud->end_xval - ud->start_xval) / 4;
   end = ud->end_xval - ( ud->end_xval - ud->start_xval) / 4;

   am_cmd_zoom_absolute(ud, start, end, userPrefs->setLogX );
}

static void
aw_zoom_out_action (GtkAction *action, UserData *ud )
{
   double start;
   double end;
   
   start = ud->start_xval - ( ud->end_xval - ud->start_xval) / 2;
   end = ud->end_xval + ( ud->end_xval - ud->start_xval) / 2;

   am_cmd_zoom_absolute(ud, start, end, userPrefs->setLogX );
}

static void
aw_zoom_cursors_action (GtkAction *action,  UserData *ud )
{
   double start;
   double end;
   
   if ( ud->cursors[0]->shown && ud->cursors[1]->shown ) {
      start = ud->cursors[0]->xval;
      end = ud->cursors[1]->xval;
  
      ud->cursors[0]->shown = ud->cursors[1]->shown = 0;
      am_cmd_zoom_absolute(ud, start, end, userPrefs->setLogX );
      
   }
}

static void
aw_zoom_x_full_action (GtkAction *action, UserData *ud )
{
   double start;
   double end;
   
   start = ud->min_xval;
   end = ud->max_xval;
   
   am_cmd_zoom_absolute(ud, start, end, userPrefs->setLogX );
}

static void
aw_zoom_y_full_action (GtkAction *action, UserData *ud )
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );

   if ( wp ) {
      wp->man_yzoom = 0;
      wp->start_yval = wp->min_yval;
      wp->end_yval = wp->max_yval;
      ap_panel_full_redraw(wp);
   } else {
      aw_redraw_all_action (action, ud );
   }
}

static void
aw_zoom_x_action (GtkAction *action,  UserData *ud)
{
   ud->srange->type = SR_X;
   ud->mouseState = M_SELRANGE_ARMED;
   ap_set_all_wp_cursors(ud, GDK_RIGHT_SIDE);
}

static void
aw_zoom_y_action (GtkAction *action,  UserData *ud)
{
   ud->srange->type = SR_Y;
   ud->mouseState = M_SELRANGE_ARMED;
   ap_set_all_wp_cursors(ud, GDK_TOP_SIDE);
}

static void
aw_zoom_xy_area_action (GtkAction *action,  UserData *ud)
{
   ud->srange->type = SR_XY;
   ud->mouseState = M_SELRANGE_ARMED;
   ap_set_all_wp_cursors(ud, GDK_TOP_LEFT_CORNER);
}

typedef struct _ToggledData {
   WavePanel *wp;
   GtkWidget *entry_ys;
   GtkWidget *entry_ye;
} ToggledData;


static void
aw_button_toggled_cb (GtkWidget *widget, ToggledData *pdata)
{
   pdata->wp->man_yzoom = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
   gtk_widget_set_sensitive (pdata->entry_ys, pdata->wp->man_yzoom);
   gtk_widget_set_sensitive (pdata->entry_ye, pdata->wp->man_yzoom);
}

static void
aw_entry_changed_cb (GtkWidget *widget, gpointer pdata)
{
   const gchar *text;
   double *val = ( double *) pdata;

   text = gtk_entry_get_text (GTK_ENTRY (widget));
   *val = str2val((char *) text);
}

static void
aw_zoom_dialog_action (GtkAction *action,  UserData *ud)
{
   WavePanel *wp = (WavePanel *) aw_get_menu_data (ud->ui, "/PanelPopup",
						    "PanelPopup-action" );
   GtkWidget *dialog;
   GtkWidget *vbox;
   gint response;
   static ToggledData *toggledData  = NULL;
   GtkWidget *frame;
   GtkWidget *x_table;
   GtkWidget *y_table;
   GtkWidget *label;
   GtkWidget *entry;
   GtkWidget *button;
   gchar *str;
   gint xpad = 5;
   gint ypad = 2;

   if ( ! wp ) {
      return;
   }
   if (! toggledData) {
      toggledData = g_new0 (ToggledData, 1);
   }
   toggledData->wp = wp;

   dialog = gtk_dialog_new_with_buttons ("Gaw axis settings",
					 GTK_WINDOW (ud->window),
					 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					 GTK_STOCK_OK,
					 GTK_RESPONSE_OK,
					 GTK_STOCK_CANCEL,
					 GTK_RESPONSE_CANCEL,
					 NULL);
   vbox = GTK_WIDGET(GTK_DIALOG (dialog)->vbox);
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 8);
   
   /* frame Global X Axis */
   frame = gtk_frame_new (" Global X Axis");
   gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, TRUE, 0);
   x_table =  gtk_table_new (4, 3, FALSE);
   gtk_container_add (GTK_CONTAINER (frame), x_table);
      
   /* frame Panel Y Axis */
   frame = gtk_frame_new (" Panel Y Axis");
   gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, TRUE, 0);
   y_table =  gtk_table_new (4, 3, FALSE);
   gtk_container_add (GTK_CONTAINER (frame), y_table);

   str = "min";
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 1, 2, 0, 1, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 1, 2, 0, 1, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );

   str = "max";
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 2, 3, 0, 1, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 2, 3, 0, 1, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   str = "Current";
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 0, 1, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, 12 + xpad, ypad );
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 0, 1, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   str = "New:";
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 0, 1, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 0, 1, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   /* X values */
   str = val2str(ud->min_xval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 1, 2, 1, 2, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   str = val2str(ud->max_xval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 2, 3, 1, 2, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );

   str = val2str(ud->start_xval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 1, 2, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   str = val2str(ud->end_xval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(x_table), label, 2, 3, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   /* Y values */
   str = val2str(wp->min_yval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 1, 2, 1, 2, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   str = val2str(wp->max_yval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 2, 3, 1, 2, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   str = val2str(wp->start_yval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 1, 2, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   str = val2str(wp->end_yval, CNV_NORMAL);
   label = gtk_label_new( str );
   gtk_table_attach (GTK_TABLE(y_table), label, 2, 3, 2, 3, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   /* new X values in entry */
   str = val2str(ud->start_xval, CNV_NORMAL);
   entry = gtk_entry_new() ;
   gtk_entry_set_text (GTK_ENTRY (entry), str);
   g_signal_connect (entry, "changed",
		     G_CALLBACK (aw_entry_changed_cb),
		     ( gpointer) &ud->start_xval);
   gtk_table_attach (GTK_TABLE(x_table), entry, 1, 2, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   str = val2str(ud->end_xval, CNV_NORMAL);
   entry = gtk_entry_new() ;
   gtk_entry_set_text (GTK_ENTRY (entry), str);
   g_signal_connect (entry, "changed",
		     G_CALLBACK (aw_entry_changed_cb),
		     ( gpointer) &ud->end_xval);
   gtk_table_attach (GTK_TABLE(x_table), entry, 2, 3, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   /* new Y values in entry */
   str = val2str(wp->start_yval, CNV_NORMAL);
   entry = gtk_entry_new() ;
   toggledData->entry_ys = entry;
   gtk_entry_set_text (GTK_ENTRY (entry), str);
   g_signal_connect (entry, "changed",
		     G_CALLBACK (aw_entry_changed_cb),
		     ( gpointer) &wp->start_yval);
   gtk_table_attach (GTK_TABLE(y_table), entry, 1, 2, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   gtk_widget_set_sensitive (entry, wp->man_yzoom);
   str = val2str(wp->end_yval, CNV_NORMAL);
   entry = gtk_entry_new() ;
   toggledData->entry_ye = entry;
   gtk_entry_set_text (GTK_ENTRY (entry), str);
   g_signal_connect (entry, "changed",
		     G_CALLBACK (aw_entry_changed_cb),
		     ( gpointer) &wp->end_yval);
   gtk_table_attach (GTK_TABLE(y_table), entry, 2, 3, 3, 4, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   gtk_widget_set_sensitive (entry, wp->man_yzoom);

   /* Full Scale button */
   str = "Full-Scale";
   button = gtk_toggle_button_new_with_label (str) ;
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), wp->man_yzoom);
   g_signal_connect (button, "toggled",
		     G_CALLBACK (aw_button_toggled_cb),
		     ( gpointer) toggledData );
   gtk_table_attach (GTK_TABLE(y_table), button, 0, 1, 0, 1, 
		     0, GTK_EXPAND|GTK_FILL, xpad, ypad );
   
   gtk_widget_show_all (vbox);
   response = gtk_dialog_run (GTK_DIALOG (dialog));

  if (response == GTK_RESPONSE_OK) {
     msg_dbg(" dialog OK");
     am_cmd_zoom_absolute(ud, ud->start_xval, ud->end_xval, userPrefs->setLogX );
     aw_redraw_all_action (action, ud );
  } else {
     wp->start_yval = wp->min_yval;
     wp->end_yval = wp->max_yval;
     wp->man_yzoom = 0;
  }

  gtk_widget_destroy (dialog);

}


static void
aw_delete_this_wave_action (GtkAction *action,  UserData *ud)
{
   msg_dbg("called");
   VwWpInfo *vwwp =  (VwWpInfo *) aw_get_menu_data (ud->ui, "/ButtonPopup",
						    "ButtonPopup-action" );
   if ( vwwp ) {
      WavePanel *wp = vwwp->wp;
      ap_remove_wave_from_panel(vwwp->wp, vwwp->vw);
      ap_panel_full_redraw(wp);
   }
}

static void
aw_delete_all_wave_action (GtkAction *action,  UserData *ud)
{
   ap_remove_all_wave_selected(ud);
}

static void
aw_reload_all_files_action (GtkAction *action,  UserData *ud )
{
   g_list_foreach(ud->wdata_list, (GFunc) ap_reload_wave_file, NULL);
}


static void
aw_var_list_create_action (GtkAction *action, DataFile *wdata)
{
   /* wlist_win is null if win has been destroyed */
   if ( wdata) {
      msg_dbg( "calling ap_create_list_win");
      ap_create_list_win (wdata->ud, wdata);
   }
}

static void
aw_script_exec_action (GtkAction *action,  UserData *ud)
{
   msg_info( "aw_script_exec_action has to be implemented...");
}

static void
aw_show_about_action (GtkAction *action,  UserData *ud)
{
   gaw_show_about_dialog (GTK_WIDGET (ud->window)); 
}

static GtkActionEntry entries[] = {
  { "FileMenu", NULL, "_File" },               /* name, stock id, label */
  { "ViewMenu", NULL, "_View" },               /* name, stock id, label */
  { "VLMenuAction", GTK_STOCK_ADD, "Variable List" },   /* name, stock id, label */
  { "PreferencesMenu", NULL, "_Preferences" }, /* name, stock id, label */
  { "HelpMenu", NULL, "_Help" },               /* name, stock id, label */
  { "Open", GTK_STOCK_OPEN,                    /* name, stock id */
    "_Open...","<control>O",                   /* label, accelerator */     
    "Open a file",                             /* tooltip */
    G_CALLBACK (aw_open_file_action) }, 
  { "Plot", GTK_STOCK_PRINT,                   /* name, stock id */
    "_Plot...", "<control>P",                  /* label, accelerator */     
    "Output dat in plot format",               /* tooltip */
    G_CALLBACK (activate_action) }, 
  { "SaveConf", GTK_STOCK_SAVE,                /* name, stock id */
    "_Save conf","<control>S",                 /* label, accelerator */     
    "Save configuration file",                 /* tooltip */
    G_CALLBACK (aw_save_config_action) },
  { "Execute", GTK_STOCK_EXECUTE,               /* name, stock id */
    "Execute script...", NULL,                  /* label, accelerator */     
    "Execute a guile script",                   /* tooltip */
    G_CALLBACK (aw_script_exec_action) },
  { "Quit", GTK_STOCK_QUIT,                    /* name, stock id */
    "_Quit", "<control>Q",                     /* label, accelerator */     
    "Quit",                                    /* tooltip */
    G_CALLBACK (activate_action) },
  { "AddPanel", GTK_STOCK_ADD,                 /* name, stock id */
    "_Add Panel", NULL,                        /* label, accelerator */     
    "Add a new panel",                         /* tooltip */
    G_CALLBACK (aw_add_panel_action) },
  { "AddPanelAbove", GTK_STOCK_ADD,            /* name, stock id */
    "_Add Panel Above", NULL,                  /* label, accelerator */     
    "Add a new panel above this one",          /* tooltip */
    G_CALLBACK (aw_add_panel_above_action) },
  { "DeletePanel", GTK_STOCK_CLEAR,            /* name, stock id */
    "_Remove Panel", "<control>D",             /* label, accelerator */     
    "Remove selected or last panel",           /* tooltip */
    G_CALLBACK (aw_remove_panel_action) },
  { "DelThisPanel", GTK_STOCK_CLEAR,           /* name, stock id */
    "_Remove this Panel", "<control>D",        /* label, accelerator */     
    "Remove this panel",                       /* tooltip */
    G_CALLBACK (aw_remove_this_panel_action) },
  { "ZoomCursors", GTK_STOCK_ZOOM_FIT,         /* name, stock id */
    "_Zoom Cursors", NULL,                     /* label, accelerator */     
    "Display space between cursors",           /* tooltip */
    G_CALLBACK (aw_zoom_cursors_action) },
  { "ZoomXFull", GTK_STOCK_ZOOM_FIT,           /* name, stock id */
    "Zoom X Full", NULL,                       /* label, accelerator */     
    "Unzoom X all",                            /* tooltip */
    G_CALLBACK (aw_zoom_x_full_action) },
  { "ZoomYFull", GTK_STOCK_ZOOM_FIT,           /* name, stock id */
    "Zoom Y Full", NULL,                       /* label, accelerator */     
    "Unzoom Y  all",                           /* tooltip */
    G_CALLBACK (aw_zoom_y_full_action) },
  { "ZoomX", GTK_STOCK_ZOOM_FIT,               /* name, stock id */
    "Zoom X...", NULL,                         /* label, accelerator */     
    "Select X range with button 1",            /* tooltip */
    G_CALLBACK (aw_zoom_x_action) },
  { "ZoomY", GTK_STOCK_ZOOM_FIT,               /* name, stock id */
    "Zoom Y...", NULL,                         /* label, accelerator */     
    "Select Y range with button 1",            /* tooltip */
    G_CALLBACK (aw_zoom_y_action) },
  { "ZoomXYarea", GTK_STOCK_ZOOM_FIT,          /* name, stock id */
    "Zoom X+Y...", NULL,                       /* label, accelerator */     
    "Select X+Y range with button 1",          /* tooltip */
    G_CALLBACK (aw_zoom_xy_area_action) },
  { "ZoomDialog", GTK_STOCK_ZOOM_FIT,          /* name, stock id */
    "Zoom Dialog...", NULL,                    /* label, accelerator */     
    "Zoom Dialog menu",                        /* tooltip */
    G_CALLBACK (aw_zoom_dialog_action) },
  { "RedrawAll", GTK_STOCK_REFRESH,            /* name, stock id */
    "_Redraw All", "<control>R",               /* label, accelerator */     
    "Redraw All panels",                       /* tooltip */
    G_CALLBACK (aw_redraw_all_action) },
  { "About", NULL,                             /* name, stock id */
    "_About", NULL,                            /* label, accelerator */     
    "About",                                   /* tooltip */  
    G_CALLBACK (aw_show_about_action) },
  { "ZoomIn", GTK_STOCK_ZOOM_IN,               /* name, stock id */
     "_Zoom In", NULL,                         /* label, accelerator */     
    "Zoom In",                                 /* tooltip */
    G_CALLBACK (aw_zoom_in_action) },
  { "ZoomOut", GTK_STOCK_ZOOM_OUT,             /* name, stock id */
     "_Zoom Out", NULL,                        /* label, accelerator */     
    "Zoom Out",                                /* tooltip */
    G_CALLBACK (aw_zoom_out_action) },
  { "DeleteWave", GTK_STOCK_DELETE,            /* name, stock id */
     "Delete Wave", NULL,                      /* label, accelerator */     
    "Delete all selected Waves",               /* tooltip */
    G_CALLBACK (aw_delete_all_wave_action) },
  { "DelThisWave", GTK_STOCK_DELETE,           /* name, stock id */
     "Delete this Wave", NULL,                 /* label, accelerator */     
    "Delete this Wave",                        /* tooltip */
    G_CALLBACK (aw_delete_this_wave_action) },
  { "ReloadAll", GTK_STOCK_FILE,               /* name, stock id */
     "Reload All", NULL,                       /* label, accelerator */     
    "Reread all waveform data files",          /* tooltip */
    G_CALLBACK (aw_reload_all_files_action) },
  { "ChangeColor", GTK_STOCK_SELECT_COLOR,     /* name, stock id */
     "Change Color...", NULL,                  /* label, accelerator */     
    "Change wave color",                       /* tooltip */
    G_CALLBACK (aw_wave_color_action) },
  { "GridColor", GTK_STOCK_SELECT_COLOR,       /* name, stock id */
     "Grid Change Color...", NULL,             /* label, accelerator */     
    "Change the grid color",                   /* tooltip */
    G_CALLBACK (aw_grid_color_action) },
  { "ExportPng", GTK_STOCK_CONVERT,            /* name, stock id */
     "Export to Image...", NULL,               /* label, accelerator */     
    "Export this wave to Image file",          /* tooltip */
    G_CALLBACK (aw_export_to_img_action) },
};
static guint n_entries = G_N_ELEMENTS (entries);

static GtkToggleActionEntry toggle_entries[] = {
  { "showXlabels", NULL,                       /* name, stock id */
     "show X labels", NULL,                    /* label, accelerator */     
    "show labels in X",                        /* tooltip */
    G_CALLBACK (aw_showXlabel_action), 
    TRUE },                                    /* is_active */
  { "showYlabels", NULL,                       /* name, stock id */
     "show Y labels", NULL,                    /* label, accelerator */     
    "show labels in Y",                        /* tooltip */
    G_CALLBACK (aw_showYlabel_action), 
    TRUE },                                    /* is_active */
  { "LogX", NULL,                              /* name, stock id */
     "log _X scale", "<control>X",             /* label, accelerator */     
    "Log scale in X",                          /* tooltip */
    G_CALLBACK (aw_logx_action), 
    FALSE },                                   /* is_active */
  { "LogY", NULL,                              /* name, stock id */
     "log _Y scale", "<control>Y",             /* label, accelerator */     
    "Set Log scale in Y for all panels",       /* tooltip */
    G_CALLBACK (aw_logy_action), 
    FALSE },                                   /* is_active */
  { "PLogY", NULL,                             /* name, stock id */
     "log _Y scale", NULL,                     /* label, accelerator */     
    "Set Log scale in Y for this panel",       /* tooltip */
    G_CALLBACK (aw_logy_pop_action), 
    FALSE },                                   /* is_active */
  { "ShowGrid", NULL,                          /* name, stock id */
     "Show Grid", NULL,                        /* label, accelerator */     
    "Show Grid in panels",                     /* tooltip */
    G_CALLBACK (aw_show_grid_action), 
    FALSE },                                   /* is_active */
  { "PshowGrid", NULL,                         /* name, stock id */
     "Show Grid", NULL,                        /* label, accelerator */     
    "Show Grid in this panel",                 /* tooltip */
    G_CALLBACK (aw_show_grid_pop_action), 
    FALSE },                                   /* is_active */
  { "Ydiff", NULL,                             /* name, stock id */
     "show Y diff measure", "<control>Y",      /* label, accelerator */     
    "Log scale in X",                          /* tooltip */
    G_CALLBACK (aw_ydiff_action), 
    FALSE },                                   /* is_active */
  { "BarStyle", GTK_STOCK_PREFERENCES,         /* name, stock id */
    "ToolBar Style Icon", NULL,                /* label, accelerator */     
    "Toggle toolbar style",                    /* tooltip */  
    G_CALLBACK (toggle_bar_style_action),
    TRUE },
};
static guint n_toggle_entries = G_N_ELEMENTS (toggle_entries);


static const gchar *ui_info = 
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Open'/>"
"      <menuitem action='Plot'/>"
"      <separator/>"
"      <menuitem action='SaveConf'/>"
"      <menuitem action='Execute'/>"
"      <separator/>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='ViewMenu'>"
"      <menuitem action='AddPanel'/>"
"      <menuitem action='ZoomCursors'/>"
"      <menuitem action='ZoomX'/>"
"      <menuitem action='ZoomY'/>"
"      <menuitem action='ZoomXFull'/>"
"      <menuitem action='ZoomYFull'/>"
"      <menuitem action='ZoomXYarea'/>"
"      <menuitem action='RedrawAll'/>"
"      <menu  action='VLMenuAction'>"
"      </menu>"
"    </menu>"
"    <menu action='PreferencesMenu'>"
"      <menuitem action='BarStyle'/>"
"      <menuitem action='showXlabels'/>"
"      <menuitem action='showYlabels'/>"
"      <menuitem action='LogX'/>"
"      <menuitem action='LogY'/>"
"      <menuitem action='Ydiff'/>"
"      <menuitem action='ShowGrid'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"  <toolbar  name='ToolBar'>"
"    <toolitem action='AddPanel'/>"
"    <toolitem action='DeletePanel'/>"
"    <separator action='Sep1'/>"
"    <toolitem action='DeleteWave'/>"
"    <toolitem action='ReloadAll'/>"
"    <separator action='Sep2'/>"
"    <toolitem action='ZoomIn'/>"
"    <toolitem action='ZoomOut'/>"
"    <toolitem action='ZoomCursors'/>"
"  </toolbar>"
"  <popup  name='PanelPopup'>"
"    <menuitem action='ZoomCursors'/>"
"    <menuitem action='ZoomX'/>"
"    <menuitem action='ZoomY'/>"
"    <menuitem action='ZoomXFull'/>"
"    <menuitem action='ZoomYFull'/>"
"    <menuitem action='ZoomXYarea'/>"
"    <menuitem action='ZoomDialog'/>"
"    <menuitem action='ZoomIn'/>"
"    <menuitem action='ZoomOut'/>"
"    <separator action='Sep1'/>"
"    <menuitem action='AddPanelAbove'/>"
"    <menuitem action='DelThisPanel'/>"
"    <menuitem action='ExportPng'/>"
"    <separator action='Sep2'/>"
"    <menuitem action='ReloadAll'/>"
"    <menuitem action='DeleteWave'/>"
"    <menuitem action='PLogY'/>"
"    <menuitem action='PshowGrid'/>"
"    <menuitem action='GridColor'/>"
"  </popup>"
"  <popup  name='ButtonPopup'>"
"    <menuitem action='DelThisWave'/>"
"    <menuitem action='ChangeColor'/>"
"    <menuitem action='ReloadAll'/>"
"  </popup>"
"</ui>";


typedef struct _ActionStatus ActionStatus;

struct _ActionStatus {
   GtkAction *action;
   GtkWidget *statusbar;
};

static void
aw_action_status_destroy (gpointer data)
{
   ActionStatus *action_status = data;
   
   g_object_unref (action_status->action);
   g_object_unref (action_status->statusbar);

   g_free (action_status);
}

static void
aw_set_tip (GtkWidget *widget)
{
   ActionStatus *data;
   gchar *tooltip;
  
   data = g_object_get_data (G_OBJECT (widget), "action-status");
  
   if (data) { 
      g_object_get (data->action, "tooltip", &tooltip, NULL);
      
      gtk_statusbar_push (GTK_STATUSBAR (data->statusbar), 0, 
			  tooltip ? tooltip : "");
      
      g_free (tooltip);
   }
}

static void
aw_unset_tip (GtkWidget *widget)
{
  ActionStatus *data;

  data = g_object_get_data (G_OBJECT (widget), "action-status");

  if (data) {
    gtk_statusbar_pop (GTK_STATUSBAR (data->statusbar), 0);
  }
}
		    
void
aw_connect_proxy (GtkUIManager *ui, GtkAction *action, GtkWidget *proxy,
	       GtkWidget *statusbar)
{
//   msg_info("called");
   if (GTK_IS_MENU_ITEM (proxy)) {
      ActionStatus *data;

      data = g_object_get_data (G_OBJECT (proxy), "action-status");
      if (data) {
	 g_object_unref (data->action);
	 g_object_unref (data->statusbar);
	 
	 data->action = g_object_ref (action);
	 data->statusbar = g_object_ref (statusbar);
      } else {
	 data = g_new0 (ActionStatus, 1);
	 
	 data->action = g_object_ref (action);
	 data->statusbar = g_object_ref (statusbar);
	 
	 g_object_set_data_full (G_OBJECT (proxy), "action-status", 
				 data, aw_action_status_destroy);
	  
	 g_signal_connect (proxy, "select",
			   G_CALLBACK (aw_set_tip), NULL);
	 g_signal_connect (proxy, "deselect",
			   G_CALLBACK (aw_unset_tip), NULL);
      }
   }
}

void 
aw_vl_menu_item_add(  UserData *ud, DataFile *wdata)
{
   gchar name[32] ;
   GtkAction *action;
   const gchar *path = "/MenuBar/ViewMenu/VLMenuAction";
   const gchar *tooltip = "reCreate variable List window";
   guint merge_id = gtk_ui_manager_new_merge_id (ud->ui);
   
   if (ud->vlactions == NULL ) {
      ud->vlactions = gtk_action_group_new ("DynamicActions");
      gtk_ui_manager_insert_action_group (ud->ui, ud->vlactions, 0);
   }
   
   sprintf(name, "varlist%d", wdata->ftag);
   action = g_object_new (GTK_TYPE_ACTION,
			  "name", name,
			  "label", wdata->label,
			  "tooltip" , tooltip,
			  "stock-id", GTK_STOCK_FILE,
			  NULL);
   gtk_action_group_add_action (ud->vlactions, action);
   g_signal_connect (action, "activate",
		     G_CALLBACK (aw_var_list_create_action), wdata);
   g_object_unref (action);
   
   gtk_ui_manager_add_ui (ud->ui, merge_id, path,  name, name,
                         GTK_UI_MANAGER_MENUITEM, FALSE );
   gtk_ui_manager_ensure_update (ud->ui);
   msg_dbg("%s", name);
}

void 
aw_update_toggle_state(GtkUIManager *ui, const gchar *path, gboolean is_active)
{
   GtkAction *action = gtk_ui_manager_get_action (ui, path);
   
   if (action) {
      if ( gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != is_active) {
	 gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), is_active);
      }
   } else {
      msg_fatal ("Can't not find action '%s'", path);
   }
}


GtkUIManager *
aw_create_ui_manager ( UserData *ud )
{
   GtkActionGroup *actions;
   GError *error = NULL;
   GtkUIManager *ui;
   GtkWidget *label;
   GtkWidget *bar;
   GtkWidget *toolitem;
   GtkWidget *table;
   GtkWidget *statusbar;
   
   actions = gtk_action_group_new ("Actions");
   gtk_action_group_add_actions (actions, entries, n_entries, ud );
   gtk_action_group_add_toggle_actions (actions, 
					toggle_entries, n_toggle_entries, 
					ud);
   /* status bar */
   statusbar = gtk_statusbar_new();
   ud->statusbar = statusbar;
   gtk_widget_show(statusbar);
   label = GTK_WIDGET(GTK_STATUSBAR (statusbar)->label);
   gtk_widget_set_name(label, "statusbar" );
   g_object_ref (statusbar); /* increment ref to avoid destruction */

   /* ui manager */
   ui = gtk_ui_manager_new ();
   userData->ui = ui;
   g_signal_connect (ud->ui, "connect-proxy",
		     G_CALLBACK (aw_connect_proxy), statusbar);
   g_object_set_data_full (G_OBJECT (ud->window), "ui-manager", ui, 
			      g_object_unref);       
   gtk_ui_manager_insert_action_group (ui, actions, 0);
   gtk_window_add_accel_group (GTK_WINDOW (ud->window), 
			       gtk_ui_manager_get_accel_group (ui));
      
   if (!gtk_ui_manager_add_ui_from_string (ui, ui_info, -1, &error)){
      msg_error ("building menus failed: %s", error->message);
      g_error_free (error);
   }

   table = gtk_table_new (1, 6, FALSE); /* 1 row 6 columns auto resize ! */
   gtk_widget_show (table);
   gtk_box_pack_start (GTK_BOX (ud->mainBox), table,
		       FALSE, FALSE, 0);

   bar = gtk_ui_manager_get_widget (ui, "/MenuBar");
   gtk_widget_show (bar);
   gtk_table_attach (GTK_TABLE (table),
		     bar, 
		     /* X direction */          /* Y direction */
		     0, 1,                      0, 1,
		     GTK_EXPAND | GTK_FILL,     0,
		     0,                         0);
   /* Create the toolbar */
   bar = gtk_ui_manager_get_widget (ui, "/ToolBar");
   userData->toolBar = bar;
   gtk_toolbar_set_tooltips (GTK_TOOLBAR (bar), TRUE);
   gtk_widget_show (bar);

   gtk_table_attach (GTK_TABLE (table),
		     bar, 
		     /* X direction */       /* Y direction */
		     0, 1,                   1, 2,
		     GTK_EXPAND | GTK_FILL,  0,
		     0,                      0);

   /* add target dnd to toolbar delete button */
   toolitem = gtk_ui_manager_get_widget (ud->ui, "/ToolBar/DeleteWave");
   ap_set_drag_dest(toolitem, NULL);

   /* modify toggle button state from userPrefs */
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/showXlabels", userPrefs->showXLabels);
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/showYlabels", userPrefs->showYLabels);
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/LogX", userPrefs->setLogX);
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/LogY", userPrefs->setLogY);
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/Ydiff", userPrefs->showYDiff);
   aw_update_toggle_state(ui, "/MenuBar/PreferencesMenu/ShowGrid", userPrefs->showGrid);

   return ui;
}

/****************************************************************/
/*
 * menu stuff for variable list window
 */

static void
aw_vl_reload_action (GtkAction *action, DataFile *wdata )
{
  ap_reload_wave_file (wdata);
}

static void
aw_vl_unload_action (GtkAction *action, DataFile *wdata )
{
   ap_delete_wave_file(NULL, wdata);
}

static void
aw_vl_export_data_action (GtkAction *action, DataFile *wdata )
{
   msg_dbg("called");
   msg_info( "aw_vl_export_data_action has to be implemented...");
}

static void
aw_vl_close_action (GtkAction *action, DataFile *wdata )
{
   msg_dbg("called");
   gtk_widget_destroy (wdata->wlist_win);
}


static GtkActionEntry vlentries[] = {
  { "FileMenu", NULL, "_File" },               /* name, stock id, label */
  { "Reload", GTK_STOCK_REFRESH,               /* name, stock id */
    "_Reload this file", NULL,                 /* label, accelerator */     
    "Reload this file",                        /* tooltip */
    G_CALLBACK (aw_vl_reload_action) },
  { "Unload", GTK_STOCK_CLEAR,                 /* name, stock id */
    "_Unload this file", NULL,                 /* label, accelerator */     
    "Unload this file",                        /* tooltip */
    G_CALLBACK (aw_vl_unload_action) },
  { "Export", GTK_STOCK_CONVERT,               /* name, stock id */
     "Export data...", NULL,                   /* label, accelerator */     
    "Export data from this file",              /* tooltip */
    G_CALLBACK (aw_vl_export_data_action) },
  { "Close", GTK_STOCK_QUIT,                    /* name, stock id */
    "_Close", NULL,                            /* label, accelerator */     
    "Close this window",                        /* tooltip */
    G_CALLBACK (aw_vl_close_action) },
};
static guint n_vlentries = G_N_ELEMENTS (vlentries);

static const gchar *vlui_info = 
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Reload'/>"
"      <menuitem action='Unload'/>"
"      <menuitem action='Export'/>"
"      <separator/>"
"      <menuitem action='Close'/>"
"    </menu>"
"  </menubar>"
"</ui>";

GtkUIManager *
aw_vl_create_ui_manager ( DataFile *wdata )
{
   GtkActionGroup *actions;
   GError *error = NULL;
   GtkUIManager *ui;
   GtkWidget *bar;
   
   actions = gtk_action_group_new ("Actions");
   gtk_action_group_add_actions (actions, vlentries, n_vlentries, wdata );

   /* ui manager */
   ui = gtk_ui_manager_new ();
   wdata->ui = ui;
   g_signal_connect ( ui, "connect-proxy",
		     G_CALLBACK (aw_connect_proxy), wdata->ud->statusbar);
   g_object_set_data_full (G_OBJECT (wdata->wlist_win), "ui-manager", ui, 
			      g_object_unref);       
   gtk_ui_manager_insert_action_group (ui, actions, 0);
   gtk_window_add_accel_group (GTK_WINDOW (wdata->wlist_win), 
			       gtk_ui_manager_get_accel_group (ui));
      
   if ( ! gtk_ui_manager_add_ui_from_string (ui, vlui_info, -1, &error)){
      msg_error ("building menus failed: %s", error->message);
      g_error_free (error);
   }

   bar = gtk_ui_manager_get_widget (ui, "/MenuBar");
   gtk_widget_show (bar);
   gtk_box_pack_start (GTK_BOX (wdata->wlist_vbox), bar,
		       FALSE, FALSE, 0);

   return ui;
}

/****************************************************************/

/*
 * GtkDialog
 */

void
aw_dialog_show ( int type, const char *text)
{
   GtkWidget *dialog;
   UserData *ud = userData;
   char buffer[64] = "";
   GtkWidget *vbox;
   GtkWidget *hbox;
   GtkWidget *image;
   GtkWidget *label;
   gchar *stock_id[] = {
      GTK_STOCK_HELP,
	 GTK_STOCK_DIALOG_INFO,
	 GTK_STOCK_DIALOG_WARNING,
	 GTK_STOCK_DIALOG_ERROR,
	 GTK_STOCK_STOP,
   };

   if ( type == MSG_T_FATAL) {
      strcpy( buffer, "Fatal - ");
   }
   strcat(buffer, "Gaw Dialog");
   
   if ( ! GTK_WIDGET_REALIZED (ud->window)) {
      fprintf(stderr,  text );
      aw_do_exit (ud->window, ud->window);
   }

   dialog = gtk_dialog_new_with_buttons (buffer,
                                         GTK_WINDOW (ud->window),
                                         GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_STOCK_OK,
                                         GTK_RESPONSE_OK,
                                         NULL);

   vbox = GTK_WIDGET(GTK_DIALOG (dialog)->vbox);
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 8);

   hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), 
		       hbox, TRUE, TRUE, 0);

   image = gtk_image_new_from_stock  ( stock_id[type],
				      GTK_ICON_SIZE_DIALOG );
   gtk_box_pack_start (GTK_BOX (hbox), 
		       image, TRUE, TRUE, 0);

   label = gtk_label_new ( text );
   gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
   gtk_label_set_max_width_chars (GTK_LABEL (label), 100);
   gtk_misc_set_padding (GTK_MISC (label), 10, 10);
   gtk_box_pack_start (GTK_BOX (hbox), 
		       label, TRUE, TRUE, 0);

   gtk_widget_show_all (vbox);

   gtk_dialog_run (GTK_DIALOG (dialog));
   gtk_widget_destroy (dialog);

   if ( type == MSG_T_FATAL && ! prog_debug ){
      aw_do_exit (ud->window, ud->window);
   }
}


/*
 * Main Window and Exit
 */

void
aw_do_exit (GtkWidget *widget, GtkWidget *window)
{
   gtk_widget_destroy (window);
   g_free (userData->prog);
   g_free (userData);
   userData = NULL;

   gtk_main_quit ();
}


void
aw_create_main_window (void)
{
   GtkWidget *window;
   GtkWidget *box1;

   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   userData->window = window;
   gtk_window_set_title (GTK_WINDOW (window), userData->mainName );
   gtk_widget_set_name (window, userData->mainName );
   gtk_widget_set_uposition (window, 20, 20);

   g_signal_connect (window, "destroy",
		     G_CALLBACK (gtk_main_quit), NULL);
   g_signal_connect (window, "delete_event",
		       G_CALLBACK (gtk_false), NULL);

   box1 = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width (GTK_CONTAINER (box1), 5);
   gtk_container_add (GTK_CONTAINER (window), box1);

   userData->mainBox = box1;
   aw_create_ui_manager (userData);
   gtk_widget_show (box1);
   /* create a global tooltips group object */
   userData->tooltips = gtk_tooltips_new();   
   
//   do_editable_cells (window);
   ap_create_win_app (userData);

   if (!GTK_WIDGET_VISIBLE (window)) {
      gtk_widget_show (window);

      gdk_window_set_hints(window->window, 0,0,
			userPrefs->min_win_width,
			userPrefs->min_win_height, 0,0,
			GDK_HINT_MIN_SIZE);
   }
   if ( userData->filename ){
      ap_load_wave_file (userData, NULL);
   }
}
   

void aw_rc_parse_gtkrc()
{
   gchar *filename;
   gchar *rc = "gaw-gtkrc";
   const gchar *home;

   filename = g_strconcat ( GAWRCDIR, "/", rc, NULL);
   gtk_rc_parse (filename);
   g_free (filename);
  
   home = g_getenv ("HOME");
   if (home != NULL) {
      filename = g_strconcat (home, "/.gaw/", rc, NULL);
      gtk_rc_parse (filename);
      g_free (filename);
  }
}

int
main (int argc, char *argv[])
{
   int i;

   srand (time (NULL));
   progname = strdup( basename(argv[0]));

//   gtk_set_locale (); /* unusefull, this is set in gtk_init */

   /* init defaults */
   if ( ! userData) {
     userData = g_new0 (UserData, 1);
   }
   msg_initlog(progname, MSG_F_NO_DATE, NULL, aw_dialog_show );
   
   userData->prog = progname ;
   userData->mainName = progname;
   userData->NWColors = 6;

   rc_init_config( userData->prog, userData->prog );

   userData->reqpanels = userPrefs->npanels; 

   aw_rc_parse_gtkrc();
   
   gtk_init (&argc, &argv);
   for ( i = 1 ; i < argc ; i++ ) {
      if (*argv[i] == '-') {
         if(strcmp(argv[i], "-d") == 0) {
	    prog_debug++ ;
#ifndef MSG_DEBUG
	    msg_info("Compile with -DMSG_DEBUG, to get debug infos.");
#endif	 
         }
      } else {
	 userData->filename = strdup(argv[i]);
      }
   }

   gdk_rgb_init ();

   aw_create_main_window ();

   gtk_main ();

   return 0;
}
