/*
 * Analog waveform viewer
 * misc stuff
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <gaw.h>

/*
 * Linear scale
 * data        |min        |val            max|
 * view            |start  |val     end|
 * pixel           |0      |x     width|
 * 
 * x_pixel = width * (val - start) / (end - start)
 * 
 * What is a log scale ?
 *   a linear space           |0      |x     width|
 *   mapped to                |Lstart |Lal    Lend|
 *   Lstart = log10 (Fstart)   : 10     hz -> 1
 *   Lend = log10 (Fend)       : 100000 hz -> 5
 *   Lval = log10 (Fval)       : 1000   hz -> 3
 */
/*
 * convert independent-variable value to pixmap X coordinate
 */
int val2x(WavePanel *wp, double val, int logs)
{
   int w = wp->drawing->allocation.width;
   double frac;

   if( logs ) {
      if(val < 0 || val < wp->start_xval) {
	 return -1;
      }

      frac = (log10(val) - log10(wp->start_xval)) /
	 (log10(wp->end_xval) - log10(wp->start_xval));
   } else {
      frac = (val - wp->start_xval) /  (wp->end_xval - wp->start_xval);
   }
   return w * frac;
}

/*
 * convert user value to pixmap y coord
 */
int val2y(WavePanel *wp, double val)
{
   double top = wp->end_yval;
   double bot = wp->start_yval;
   int h = wp->drawing->allocation.height;
   double frac;
        
   if (wp->logy) {
      if(bot < 0 || val < 0) {
	 return -1;
      }

      frac = ( log10(val) - log10(bot)) / ( log10(top) - log10(bot));
   } else {
      frac = (val - bot ) / (top - bot);
   }

   return h - ((h - 6) * frac) - 3;
}

/*
 * convert pixmap X coordinate to user independent-variable value
 */
double x2val(WavePanel *wp, int x, int logs)
{
   int w = wp->drawing->allocation.width;
   double frac = (double)x / (double)w;

   if (logs) {
      double a;
      a = frac * ( log10(wp->end_xval) - log10(wp->start_xval)) 
	 + log10(wp->start_xval);
      return pow(10, a);
   } else {
      return frac * (wp->end_xval - wp->start_xval) 
	 + wp->start_xval;
   }
}

/*
 * convert pixmap Y coordinate to user dependent-variable value
 */
double y2val(WavePanel *wp, int y)
{
   int h = wp->drawing->allocation.height;
   double frac = - (double) (y - h + 3) / (double) (h - 6);
        
   if( wp->logy ) {
      double a;
      a = frac * (log10(wp->end_yval) - log10(wp->start_yval)) 
	 + log10(wp->start_yval);
      return pow(10, a);
   } else {
      return frac * (wp->end_yval - wp->start_yval) + wp->start_yval;
   }
}

/*
 * reset the x zoom scale of all panels
 */
gint
am_cmd_zoom_absolute(UserData *ud, double start, double end, int logs )
{
   double scroll_start, scroll_end;

   msg_dbg("start %f  end %f", start, end);
   /* set starting drawn x-value */
   if (start <= end) {
      ud->start_xval = start;
      ud->end_xval = end;
   } else {
      ud->start_xval = end;
      ud->end_xval = start;
   }

   if (ud->start_xval < ud->min_xval) {
      ud->start_xval = ud->min_xval;
   }
   if (ud->end_xval > ud->max_xval) {
      ud->end_xval = ud->max_xval;
   }

   /*
    * Scroll bar always goes from zero to one.
    * Perform an appropriate transform based on lin/log
    */
   if ( ! logs ) {
      scroll_start = ( ud->start_xval - ud->min_xval ) /
	 ( ud->max_xval - ud->min_xval );
      scroll_end   = ( ud->end_xval   - ud->min_xval ) /
	 ( ud->max_xval - ud->min_xval );
   } else {
      scroll_start = log( ud->start_xval / ud->min_xval ) / 
	 log( ud->max_xval / ud->min_xval );
      scroll_end   = log( ud->end_xval  / ud->min_xval ) / 
	 log( ud->max_xval / ud->min_xval );
   }
   ud->xadj->page_size = fabs( scroll_end - scroll_start );

   ud->xadj->page_increment = ud->xadj->page_size / 2;
   ud->xadj->step_increment = ud->xadj->page_size / 100;
   ud->xadj->value = scroll_start;
   ud->xadj->lower = 0.0;
   ud->xadj->upper = 1.0;

   gtk_adjustment_changed (ud->xadj);
   gtk_adjustment_value_changed (ud->xadj);

   return 0;
}

/*
 * convert double value to printable text.
 * Two modes:
 *   0: using suffixes
 *      We always try to print 4 significant figures, with one nonzero digit
 *      to the left of the decimal point.
 *      maximum field width: 7 characters
 *
 *   1: use scientific notation, printf %g format.
 *      maximum field width appears to be 10 characters
 */

char *val2str(double val, int mode)
{
   static char buf[64];
   double aval = fabs(val);
   double sval, asval;
   char suffix;
   int ddigits;

   if (mode == CNV_SCIENTIFIC) {
      sprintf(buf, "% .5g", val);
   } else {
      if(1e12 <= aval) {
	 suffix = 'T';
	 sval = val / 1e12;
      } else if(1e9 <= aval && aval < 1e12) {
	 suffix = 'G';
	 sval = val / 1e9;
      } else if(1e6 <= aval && aval < 1e9) {
	 suffix = 'M';
	 sval = val / 1e6;
      } else if(1e3 <= aval && aval < 1e6) {
	 suffix = 'K';
	 sval = val / 1000;
      } else if(1e-3 <= aval && aval < 1) {
	 suffix = 'm';
	 sval = val * 1000;
      } else if(1e-6 <= aval && aval < 1e-3) {
	 suffix = 'u';
	 sval = val * 1e6;
      } else if(1e-9 <= aval && aval < 1e-6) {
	 suffix = 'n';
	 sval = val * 1e9;
      } else if(1e-12 <= aval && aval < 1e-9) {
	 suffix = 'p';
	 sval = val * 1e12;
      } else if(1e-15 <= aval && aval < 1e-12) {
	 suffix = 'f';
	 sval = val * 1e15;
      } else if(DBL_EPSILON < aval && aval < 1e-15) {
	 suffix = 'a';
	 sval = val * 1e18;
      } else {
	 suffix = ' ';
	 sval = val;
      }
      asval = fabs(sval);
      if (1.0 <=  asval && asval < 10.0) { 
	 ddigits = 4;
      } else if (10.0 <=  asval && asval < 100.0) { 
	 ddigits = 3;
      } else { 
	 ddigits = 2;
      }
      sprintf(buf, "% .*f%c", ddigits, sval, suffix);
   }       
   return buf;
}

/*
 *  convert a string to double 
 *  given a string containing a number, possibly with a spice-syntax suffix,
 *  return the value
 */
double str2val(char *s)
{
   double val, mul;
   char *ep;
   val = strtod(s, &ep);

   if(ep && ep != s && *ep) {
      switch(*ep) {
       case 'T':
	 mul = 1e12;
	 break;
       case 'G':
	 mul = 1e9;
	 break;
       case 'M':
	 mul = 1e6;
	 break;
       case 'k':
       case 'K':
	 mul = 1e3;
	 break;
       case 'm':
	 mul = 1e-3;
	 break;
       case 'u':
	 mul = 1e-6;
	 break;
       case 'n':
	 mul = 1e-9;
	 break;
       case 'p':
	 mul = 1e-12;
	 break;
       case 'f':
	 mul = 1e-15;
	 break;
       case 'a':
	 mul = 1e-18;
	 break;
       default:
	 mul = 1;
	 break;
      }
      return val * mul;
   } else {
      return val;
   }
}


/*
 * called with g_list_foreach to update a WavePanel
 * from all of its  VisibleWaves.
 */
static void
am_min_max_vw_wp_update(gpointer p, gpointer d)
{
   VisibleWave *vw = (VisibleWave *) p;
   WavePanel *wp = (WavePanel *) d;
   WDataSet *dataset = vw->var->wtable->iv->wds;

   /* MIN(a, b)     ((a) < (b) ? (a) : (b)) */
   /* MAX(a, b)     ((a) > (b) ? (a) : (b)) */
   wp->min_xval = MIN( dataset->min, wp->min_xval);
   wp->max_xval = MAX( dataset->max, wp->max_xval);

   wp->min_yval = MIN( vw->var->wds[0].min, wp->min_yval);
   wp->max_yval = MAX( vw->var->wds[0].max, wp->max_yval);
}

/*
 * wavepanel_update_data
 *   update wavepanel values that sumarize things over all of the 
 *   VisibleWaves in the panel.
 */
void
am_wavepanel_update_data(WavePanel *wp)
{
   wp->min_xval = G_MAXDOUBLE;
   wp->max_xval = -G_MAXDOUBLE;
   wp->min_yval = G_MAXDOUBLE;
   wp->max_yval = -G_MAXDOUBLE;
   /* set wp min, maw x, y values from min, max in datasets */
   g_list_foreach(wp->vwlist, am_min_max_vw_wp_update, (gpointer) wp);

   /*
    * set to something reasonable if they didn't change,
    * like if the panel was empty
    */
   if (wp->min_xval == G_MAXDOUBLE) {
      wp->min_xval = 0.0; /* wtable->min_xval; */
   }
   if (wp->max_xval == - G_MAXDOUBLE) {
      wp->max_xval = 0.0; /* wtable->max_xval; */
   }
   if (wp->min_yval == G_MAXDOUBLE) {
      wp->min_yval = 0.0;
   }
   if (wp->max_yval == - G_MAXDOUBLE) {
      wp->max_yval = 1.0;
   }
   if (wp->man_yzoom == 0) {
      wp->start_yval = wp->min_yval;
      wp->end_yval = wp->max_yval;
   }
   /* zero height? set to +- 0.1%  so a line is visible in the center */
   if ((wp->end_yval - wp->start_yval) < DBL_EPSILON) {
      wp->end_yval *= 1.001;
      wp->start_yval *= 0.999;
      /* still zero?  maybe there's a waveform that is stuck at 0.000 */
      if ((wp->end_yval - wp->start_yval) < DBL_EPSILON) {
	 wp->end_yval += 1e-6;
	 wp->start_yval -= 1e-6;
      }
   }

   /*
    * if start & end were the same, try updating them
    * -- this probably isn't quite right.
    */
//   if (fabs(wp->end_xval - wp->start_xval) < DBL_EPSILON) {
   wp->start_xval = wp->min_xval;
   wp->end_xval = wp->max_xval;
//   }
/*
   if (wp->start_xval < wp->min_xval) {
      wp->start_xval = wp->min_xval;
   }
   if (wp->end_xval > wp->max_xval) {
      wp->end_xval = wp->max_xval;
   }
 */
}

   
/* 
 * Update parameters in wavetable that depend on all panels
 */
void
am_wavetable_update_data(UserData *ud)
{
   int i;
   WavePanel *wp;
   double old_min_x, old_max_x;
   old_min_x = ud->min_xval;
   old_max_x = ud->max_xval;

   ud->min_xval = G_MAXDOUBLE;
   ud->max_xval = - G_MAXDOUBLE;
   for (i = 0 ; i < ud->npanels ; i++) {
      wp = ud->panels[i];
      if (wp == NULL) {
	 continue; /* deleted panel */
      }
      if(wp->vwlist == NULL) {
	 continue; /* no waves? min/max for panel are bogus */
      }

      ud->min_xval = MIN( wp->min_xval, ud->min_xval);
      ud->max_xval = MAX( wp->max_xval, ud->max_xval);
   }
   /* still nothing? set back to zero */
   if (ud->min_xval == G_MAXDOUBLE) {
      ud->min_xval = 0;
   }
   if (ud->max_xval == - G_MAXDOUBLE) {
      ud->max_xval = 0;
   }

   /*
    * if start & end were the same or out of range, 
    * just zoom-full so we can see somthing.
    */
   if ((fabs(ud->end_xval - ud->start_xval) < DBL_EPSILON ||
	ud->start_xval < ud->min_xval ||
	ud->end_xval > ud->max_xval) && ud->xadj != NULL) {
      ud->suppress_redraw = 1;
      am_cmd_zoom_absolute(ud, ud->min_xval, ud->max_xval, userPrefs->setLogX );
      ud->suppress_redraw = 0;
   } else if((ud->min_xval != old_min_x || ud->max_xval != old_max_x) &&
	     ud->xadj != NULL ) {
      /*
       * min/max changed, might have added first (or removed last)
       * wave from a file with different range.
       * try to keep start/end same, but make them sane if needed.
       * then update scrollbar.
       */
      am_cmd_zoom_absolute( ud, ud->min_xval, ud->max_xval, userPrefs->setLogX );
   }
   /*
    * propagate zoom to panels.  Someday: flags and UI to allow
    * "locking" selected panels so they don't zoom/scroll
    */
   for (i = 0 ; i < ud->npanels; i++) {
      wp = ud->panels[i];
      wp->start_xval = ud->start_xval;
      wp->end_xval = ud->end_xval;
   }
}


/*
 * Routines to insert and delete rows in a GtkTable.
 * The table is assumed to have no children that span more than one
 * row: odd displays may result if this is not true.
 *
 * Rows are numbered by the value of the top_attach field;
 * a table with 2 rows has rows numbered 0 and 1, etc.
 */


static GtkTableChild *
am_collect_children(GtkTable *table, int row, int *nchp)
{
   GtkTableChild *row_children;
   GtkTableChild *table_child;
   int nch;
   GList *list;

   row_children = g_new0(GtkTableChild, table->ncols);
   nch = 0;

   /*
    * locate all of the relevant children, and add an extra reference
    * to them
    */
   for (list = table->children; list; list = list->next) {
      table_child = list->data;

      if (table_child->top_attach == row) {
	 if (nch == table->ncols) {
	    g_error("collect_children: too many children on row %d\n", row);
	    abort();
	 }
	 row_children[nch++] = *table_child;
	 gtk_widget_ref(table_child->widget);
      }
   }
   if(nchp)
      *nchp = nch;
   return row_children;
}

static void
am_reattach_children(GtkTable *table, GtkTableChild *row_children,
		     int nch,  int row) 
{
   int i;
   GtkAttachOptions xoptions;
   GtkAttachOptions yoptions;

   /* add in new position and unref */
   for (i = 0; i < nch; i++) {
      xoptions = row_children[i].xexpand 
	 | row_children[i].xshrink << 1
	 | row_children[i].xfill << 2;
      yoptions = row_children[i].yexpand 
	 | row_children[i].yshrink << 1
	 | row_children[i].yfill << 2;

      gtk_table_attach(table, row_children[i].widget,
		       row_children[i].left_attach,
		       row_children[i].right_attach,
		       row,
		       row + (row_children[i].bottom_attach -
			      row_children[i].top_attach),
		       xoptions,
		       yoptions,
		       row_children[i].xpadding,
		       row_children[i].ypadding);
      
      gtk_widget_unref(row_children[i].widget);
   }
}



/*
 * move a table row to another row position.
 */
void
am_gtk_table_move_row(GtkTable *table, int from, int to)
{
   GtkTableChild *row_children;
   int nch;
   int i;
   GtkTableChild *table_child;
   GtkAttachOptions xoptions;
   GtkAttachOptions yoptions;
   GList *list;

/* printf("move row %d to %d: ", from, to); */
   row_children = g_new(GtkTableChild, table->ncols);
   nch = 0;

   /* locate all of the relevant children, and add an extra reference
    to them */
   for (list = table->children; list; list = list->next) {
      table_child = list->data;
      
      if(table_child->top_attach == from) {
	 if(nch == table->ncols) {
	    g_error("gtk_table_move_row: too many children on row %d\n", from);
	    abort();
	 }
	 row_children[nch++] = *table_child;
	 gtk_widget_ref(table_child->widget);
      }
   }

   /* remove from old position */
   for(i = 0; i < nch; i++) {
      gtk_container_remove(GTK_CONTAINER(table), row_children[i].widget);
   }
   /* add in new position and unref */
   for(i = 0; i < nch; i++) {
      xoptions = row_children[i].xexpand 
	 | row_children[i].xshrink<<1
	 | row_children[i].xfill<<2;
      yoptions = row_children[i].yexpand 
	 | row_children[i].yshrink<<1
	 | row_children[i].yfill<<2;
      
      gtk_table_attach(table, row_children[i].widget,
		       row_children[i].left_attach,
		       row_children[i].right_attach,
		       to,
		       to + (row_children[i].bottom_attach -
			     row_children[i].top_attach),
		       xoptions,
		       yoptions,
		       row_children[i].xpadding,
		       row_children[i].ypadding);
      
      gtk_widget_unref(row_children[i].widget);
   }
/*	printf("%d widgets moved\n", nch); */
	
   g_free(row_children);
}

/*
 * Insert a new empty row into a GtkTable just before the specified row,
 * moving subsequent rows down.
 *	- resize the table one larger
 *	- move rows down to put the hole where it belongs
 */
void
am_gtk_table_insert_row(GtkWidget *widget, int row)
{
   GtkTable *table;
   int old_nrows, old_ncols;
   int i;
   table = GTK_TABLE(widget);
   old_nrows = table->nrows;
   old_ncols = table->ncols;
   
   gtk_table_resize(table, old_nrows+1, old_ncols);
	
   for(i = old_nrows; i > row; i--) {
      am_gtk_table_move_row(table, i-1, i);
   }
}

/*
 * Delete the indicated row of a GtkTable, moving subsequent rows up.
 *	- remove the specified row
 *	- move other rows
 *	- resize the table one smaller
 */
void
am_gtk_table_delete_row(GtkWidget *widget, int row)
{
   GtkTable *table;
   GtkTableChild *table_child;
   GList *list;
   GList *del_list = NULL;
   int old_nrows, old_ncols;
   int i;
   table = GTK_TABLE(widget);
   old_nrows = table->nrows;
   old_ncols = table->ncols;

   /*
    * remove widgets from the old row 
    * can't delete while walking the list; must make a temporary list
    */
   for (list = table->children; list; list = list->next) {
      table_child = list->data;
      if( table_child->top_attach == row) {
	 del_list = g_list_append(del_list, table_child->widget);
      }
   }

   while (del_list) {
      GtkWidget *child;
      child = del_list->data;
      /* printf("remove child 0x%lx\n", child); */
      gtk_container_remove(GTK_CONTAINER(table), child);
      del_list = g_list_remove(del_list, child);
   }

   for (i = row; i < old_nrows - 1; i++) {
      am_gtk_table_move_row(table, i + 1, i);
   }

   gtk_table_resize(table, old_nrows - 1, old_ncols);
}

/*
 * Reorder a table, moving row FROM to position TO, and shifting
 * all rows in between up or down one row.
 * given this arrangement:
0 A
1 B
2 C
3 D
rotate_rows(, 2, 0) produces this result:  ("up" case)
0 C
1 A
2 B
3 D
while applying rotate_rows(, 1, 3) to the original produces: ("down" case)
0 A
1 C
2 D
3 B
 */
void
am_gtk_table_rotate_rows(GtkWidget *widget, int from, int to)
{
   GtkTable *table = GTK_TABLE(widget);
   int down;  /* does the named "jumping" row move up or down? */
   GtkTableChild *moving_children;
   int nch, i;

   if (to == from) {
      return;
   } else if (to > from) {
      down = 1;
   } else {
      down = 0;
   }

   /* collect the children in the row that's jumping */
   moving_children = am_collect_children(table, from, &nch);
   /* remove them from their old row */
   for (i = 0; i < nch; i++) {
      gtk_container_remove(GTK_CONTAINER(table), moving_children[i].widget);
   }

   /* now move the other ones one position.
    * if the "jumping" row moves down, thers move up, etc. */
   if(down) {
      for(i = from; i < to; i++) {
	 am_gtk_table_move_row(table, i + 1, i);
      }
   } else { /* up*/
      for(i = from; i > to; i--) {
	 am_gtk_table_move_row(table, i - 1, i);
      }
   }
   
   /* finally, put those children back down in the vacated target row */
   am_reattach_children(table, moving_children, nch, to);
   g_free(moving_children);
}

/*
 * given a pointer to a child widget of a table, 
 * return the (top_attach) row number it is currently on.
 */
int
am_gtk_table_get_child_row(GtkWidget *widget, GtkWidget *child)
{
   GtkTable *table;
   GtkTableChild *table_child;
   GList *list;
   table = GTK_TABLE(widget);

   for (list = table->children; list; list = list->next) {
      table_child = list->data;
		
      if (table_child->widget == child) {
	 return table_child->top_attach;
      }
   }
   return -1;
}
